from typing import Any, Callable, Iterable, List, Optional, Union

import tempfile

import numpy as np
import pandas as pd
import yaml

from giskard.client.python_utils import warning
from giskard.core.core import ModelMeta, ModelType, SupportedModelTypes
from giskard.core.validation import configured_validate_arguments, validate_is_pandasdataframe
from giskard.datasets.base import Dataset
from giskard.models.base import BaseModel, WrapperModel
from giskard.registry.slicing_function import SlicingFunction

from ..utils import fullname
from ..utils.analytics_collector import analytics, get_dataset_properties, get_model_properties


@configured_validate_arguments
def validate_model(model: BaseModel, validate_ds: Optional[Dataset] = None, print_validation_message: bool = True):
    try:
        _do_validate_model(model, validate_ds)
    except (ValueError, TypeError) as err:
        _track_validation_error(err, model, validate_ds)
        raise err

    # TODO: switch to logger
    if print_validation_message:
        print("Your model is successfully validated.")


def _track_validation_error(err, model, dataset):
    properties = {"error": str(err), "error_class": fullname(err)}
    properties.update(get_model_properties(model))
    properties.update(get_dataset_properties(dataset))

    analytics.track("validate_model:failed", properties)


def _do_validate_model(model: BaseModel, validate_ds: Optional[Dataset] = None):
    model_type = model.model_type

    if isinstance(model, WrapperModel) and model.data_preprocessing_function is not None:
        validate_data_preprocessing_function(model.data_preprocessing_function)

    if isinstance(model, WrapperModel) and model.model_postprocessing_function is not None:
        validate_model_postprocessing_function(model.model_postprocessing_function)

    validate_classification_labels(model.classification_labels, model_type)

    if model.is_classification:
        validate_classification_threshold_label(model.classification_labels, model.classification_threshold)

    assert model.feature_names is None or isinstance(
        model.feature_names, list
    ), "Invalid feature_names parameter. Please provide the feature names as a list."

    if validate_ds is not None:
        validate_is_pandasdataframe(validate_ds.df)
        validate_features(feature_names=model.feature_names, validate_df=validate_ds.df)

        if model.is_regression:
            validate_model_execution(model, validate_ds)
        elif model.is_text_generation:
            validate_model_execution(model, validate_ds, False)
        elif model.is_classification and validate_ds.target is not None:
            target_values = validate_ds.df[validate_ds.target].unique()
            validate_label_with_target(model.name, model.classification_labels, target_values, validate_ds.target)
            validate_model_execution(model, validate_ds)
        else:  # Classification with target = None
            validate_model_execution(model, validate_ds)

        if model.model_type == SupportedModelTypes.CLASSIFICATION and validate_ds.target is not None:
            validate_order_classifcation_labels(model, validate_ds)


@configured_validate_arguments
def validate_model_execution(model: BaseModel, dataset: Dataset, deterministic: bool = True) -> None:
    # testing multiple entries
    validation_size = min(len(dataset), 10)
    validation_ds = dataset.slice(SlicingFunction(lambda x: x.sample(validation_size), row_level=False))
    error_message = (
        "Invalid prediction_function.\nBefore using Giskard, please make sure that "
        "'prediction_function(df[feature_names].head())' does not return an error message."
    )
    try:
        prediction = model.predict(validation_ds)
    except Exception as e:
        features = model.feature_names if model.feature_names is not None else validation_ds.df.columns
        number_of_features = len(features)

        # Some models (mostly sklearn) expect a 1-dimensional ndarray or pd.Series as input in the case they're
        # trained with 1 feature. Here we try to detect in case a user defines their prediction function using
        # model.predict_proba(df) (which would break) instead of model.predict_proba(df.feature)
        one_dimension_case = (
            isinstance(e, IndexError)
            and "index 1 is out of bounds for axis 0 with size 1" in str(e)
            and number_of_features == 1
        )

        if not one_dimension_case:
            raise ValueError(error_message) from e

        feature = features[0]
        one_dimension_error_message = (
            "Your model returned an error when we passed a 'pandas.Dataframe' as input. "
            f"Try to use a one-dimensional input like 'df.{feature}' "
            "inside your prediction function."
        )
        raise ValueError(one_dimension_error_message) from e

    # testing one entry
    validation_size = min(len(dataset), 1)
    validation_ds_1 = dataset.slice(SlicingFunction(lambda x: x.sample(validation_size), row_level=False))
    try:
        model.predict(validation_ds_1)
    except Exception as e:
        raise ValueError(
            error_message + " Hint: Make sure that you are not fitting any preprocessor inside your model or prediction"
            " function."
        ) from e

    if deterministic:
        validate_deterministic_model(model, validation_ds, prediction)
    validate_prediction_output(validation_ds, model.model_type, prediction.raw)
    if model.is_classification:
        validate_classification_prediction(model.classification_labels, prediction.raw)


@configured_validate_arguments
def validate_deterministic_model(model: BaseModel, validate_ds: Dataset, prev_prediction):
    """Asserts if the model is deterministic by asserting previous and current prediction on same data are same"""
    new_prediction = model.predict(validate_ds)

    if not np.allclose(prev_prediction.raw, new_prediction.raw):
        warning(
            "Model is stochastic and not deterministic. Prediction function returns different results"
            "after being invoked for the same data multiple times."
        )


@configured_validate_arguments
def validate_model_loading_and_saving(model: BaseModel):
    """Validates if the model can be serialised and deserialised from local disk"""
    try:
        with tempfile.TemporaryDirectory(prefix="giskard-model-") as f:
            model.save(f)

            with open(f + "/giskard-model-meta.yaml", encoding="utf-8") as yaml_f:
                saved_meta = yaml.load(yaml_f, Loader=yaml.Loader)

            meta = ModelMeta(
                name=saved_meta["name"],
                description=saved_meta["description"],
                model_type=SupportedModelTypes[saved_meta["model_type"]],
                feature_names=saved_meta["feature_names"],
                classification_labels=saved_meta["classification_labels"],
                classification_threshold=saved_meta["threshold"],
                loader_module=saved_meta["loader_module"],
                loader_class=saved_meta["loader_class"],
            )

            clazz = BaseModel.determine_model_class(meta, f)

            constructor_params = meta.__dict__
            del constructor_params["loader_module"]
            del constructor_params["loader_class"]

            loaded_model = clazz.load(f, **constructor_params)

            return loaded_model

    except Exception as e:
        raise ValueError("Failed to validate model saving and loading from local disk") from e


@configured_validate_arguments
def validate_data_preprocessing_function(f: Callable[[pd.DataFrame], Any]):
    if not callable(f):
        raise ValueError(f"Invalid data_preprocessing_function parameter: {f}. Please specify Python function.")


@configured_validate_arguments
def validate_model_postprocessing_function(f: Callable[[Any], Any]):
    if not callable(f):
        raise ValueError(f"Invalid model_postprocessing_function parameter: {f}. Please specify Python function.")


@configured_validate_arguments
def validate_model_type(model_type: ModelType):
    if model_type not in {task.value for task in SupportedModelTypes}:
        raise ValueError(
            f"Invalid model_type parameter: {model_type}. "
            + f"Please choose one of {[task.value for task in SupportedModelTypes]}."
        )


@configured_validate_arguments
def validate_classification_labels(classification_labels: Union[np.ndarray, List, None], model_type: ModelType):
    if model_type == SupportedModelTypes.CLASSIFICATION:
        if classification_labels is not None and isinstance(classification_labels, Iterable):
            if len(classification_labels) <= 1:
                raise ValueError(
                    f"Invalid classification_labels parameter: {classification_labels}. "
                    f"Please specify more than 1 label."
                )
        else:
            raise ValueError(
                f"Invalid classification_labels parameter: {classification_labels}. "
                f"Please specify valid list of strings."
            )

    if (
        model_type == SupportedModelTypes.REGRESSION or model_type == SupportedModelTypes.TEXT_GENERATION
    ) and classification_labels is not None:
        warning("'classification_labels' parameter is ignored for regression model")


@configured_validate_arguments
def validate_features(feature_names: Optional[List[str]] = None, validate_df: Optional[pd.DataFrame] = None):
    if (
        feature_names is not None
        and validate_df is not None
        and not set(feature_names).issubset(set(validate_df.columns))
    ):
        missing_feature_names = set(feature_names) - set(validate_df.columns)
        raise ValueError(f"Value mentioned in feature_names is not available in validate_df: {missing_feature_names} ")


@configured_validate_arguments
def validate_classification_threshold_label(
    classification_labels: Union[np.ndarray, List, None], classification_threshold: float = None
):
    if classification_labels is None:
        raise ValueError("Missing classification_labels parameter for classification model.")
    if classification_threshold is not None and not isinstance(classification_threshold, (int, float)):
        raise ValueError(
            f"Invalid classification_threshold parameter: {classification_threshold}. Please specify valid number."
        )

    if classification_threshold is not None:
        if not np.isclose(classification_threshold, 0.5, rtol=1e-09, atol=1e-09) and len(classification_labels) != 2:
            raise ValueError(
                f"Invalid classification_threshold parameter: {classification_threshold} value is applicable "
                f"only for binary classification. "
            )


@configured_validate_arguments
def validate_label_with_target(
    model_name: str,
    classification_labels: Union[np.ndarray, List, None],
    target_values: Union[np.ndarray, List, None] = None,
    target_name: str = None,
):
    if target_values is not None:
        to_append = " of the model: " + model_name if model_name else ""
        target_values = list(target_values)
        if not set(target_values).issubset(set(classification_labels)):
            invalid_target_values = set(target_values) - set(classification_labels)
            raise ValueError(
                f'Values {invalid_target_values} in "{target_name}" column are not declared in '
                f"classification_labels parameter {classification_labels}" + to_append
            )


@configured_validate_arguments
def validate_prediction_output(ds: Dataset, model_type: ModelType, prediction):
    assert len(ds.df) == len(prediction), (
        f"Number of rows ({len(ds.df)}) of dataset provided does not match with the "
        f"number of rows ({len(prediction)}) of model.predict output"
    )
    if isinstance(prediction, np.ndarray) or isinstance(prediction, list):
        if model_type == SupportedModelTypes.CLASSIFICATION:
            if not any(isinstance(y, (np.floating, float)) for x in prediction for y in x):
                raise ValueError("Model prediction should return float values ")
        if model_type == SupportedModelTypes.REGRESSION:
            if not any(isinstance(x, (np.floating, float)) for x in prediction):
                raise ValueError("Model prediction should return float values ")
        if model_type == SupportedModelTypes.TEXT_GENERATION:
            if not any(isinstance(x, str) for x in prediction):
                raise ValueError("Model prediction should return string values ")
    else:
        raise ValueError("Model should return numpy array or a list")


@configured_validate_arguments
def validate_classification_prediction(classification_labels: Union[np.ndarray, List, None], prediction):
    if not np.all(np.logical_and(prediction >= 0, prediction <= 1)):
        warning(
            "Output of model.predict returns values out of range [0,1]. "
            "The output of Multiclass and Binary classifications should be within the range [0,1]"
        )
    if not np.all(np.isclose(np.sum(prediction, axis=1), 1, atol=0.0000001)):
        warning(
            "Sum of output values of model.predict is not equal to 1."
            " For Multiclass and Binary classifications, the sum of probabilities should be 1"
        )
    if prediction.shape[1] != len(classification_labels):
        raise ValueError("Prediction output label shape and classification_labels shape do not match")
    if prediction.shape[1] != len(classification_labels):
        raise ValueError("Prediction output label shape and classification_labels shape do not match")


def validate_order_classifcation_labels(model, dataset):
    from sklearn.metrics import balanced_accuracy_score

    y_true = dataset.df[dataset.target]
    y_pred = model.predict(dataset).prediction
    balanced_accuracy = balanced_accuracy_score(y_true, y_pred)
    num_classes = len(model.classification_labels)

    if balanced_accuracy <= 1 / num_classes:
        warning(
            f"The balanced accuracy of your model is very low ({round(balanced_accuracy, 2)}). "
            "Make sure you have not inverted the order of the 'classification_labels' when you created "
            "the Giskard Model."
        )



from typing import Dict, Optional, Sequence

import json
import logging
from abc import ABC, abstractmethod

import pandas as pd
from pydantic import BaseModel

from ...datasets.base import Dataset
from ..client import LLMClient, get_default_client
from ..client.base import ChatMessage
from ..errors import LLMGenerationError

logger = logging.getLogger("giskard.llm")


class BaseGenerator(ABC):
    @abstractmethod
    def generate_dataset(self, model, num_samples=10, column_types=None) -> Dataset:
        ...


class _BaseLLMGenerator(BaseGenerator, ABC):
    _default_temperature = 0.5
    _output_key = "inputs"

    def __init__(
        self,
        languages: Optional[Sequence[str]] = None,
        llm_temperature: Optional[float] = None,
        llm_client: LLMClient = None,
        llm_seed: int = 1729,
    ):
        self.languages = languages or ["en"]
        self.llm_temperature = llm_temperature if llm_temperature is not None else self._default_temperature
        self.llm_client = llm_client or get_default_client()
        self.llm_seed = llm_seed

    def generate_dataset(self, model: BaseModel, num_samples: int = 10, column_types: Dict = None) -> Dataset:
        """Generates a test dataset for the model.

        Parameters
        ----------
        model : BaseModel
            The model to generate a test dataset for.
        num_samples : int
            The number of samples to generate, by default 10.
        column_types : dict, optional
            The column types for the generated datasets. (Default value = None)

        Returns
        -------
        Dataset
            The generated dataset.

        Raises
        ------
        LLMGenerationError
            If the generation fails.
        """
        messages = self._format_messages(model, num_samples, column_types)

        out = self.llm_client.complete(
            messages=messages,
            temperature=self.llm_temperature,
            caller_id=self.__class__.__name__,
            seed=self.llm_seed,
            format="json_object",
        )

        generated = self._parse_output(out)

        dataset = Dataset(
            df=pd.DataFrame(generated),
            name=self._make_dataset_name(model),
            validation=False,
            column_types=column_types,
        )

        return dataset

    def _parse_output(self, raw_output: ChatMessage):
        try:
            data = json.loads(raw_output.content)
            if self._output_key:
                data = data[self._output_key]
        except (json.JSONDecodeError, KeyError) as err:
            logger.error("Generator output parse error, got raw output: %s", raw_output.content)
            raise LLMGenerationError("Could not parse generated data") from err
        return data

    def _make_dataset_name(self, model: BaseModel):
        return f"Synthetic Test Dataset for {model.name}"

    @abstractmethod
    def _format_messages(self, model: BaseModel, num_samples: int):
        ...


class LLMBasedDataGenerator(_BaseLLMGenerator):
    def __init__(
        self,
        prompt: str,
        prefix_messages: Optional[Sequence[ChatMessage]] = None,
        languages: Optional[Sequence[str]] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.prompt = prompt
        self.prefix_messages = prefix_messages or []
        self.languages = languages or ["en"]

    def _format_messages(
        self, model: BaseModel, num_samples: int, column_types: Optional[Dict] = None
    ) -> Sequence[ChatMessage]:
        prompt = self.prompt.format(model=model, num_samples=num_samples, languages=", ".join(self.languages))
        return self.prefix_messages + [ChatMessage(role="user", content=prompt)]



"""Various input/output utility functions"""

from typing import Any, Optional

import json
import re
from io import BytesIO

import cloudpickle
import numpy as np
import pandas as pd
from zstandard import ZstdCompressor, ZstdDecompressor

COMPRESSION_MAX_OUTPUT_SIZE = 10**9  # 1GB


def pickle_dumps(variable: object) -> bytes:
    pickle: bytes = cloudpickle.dumps(variable)
    return pickle


def pickle_loads(dumped_pickle: bytes) -> Any:
    return cloudpickle.loads(dumped_pickle)


def load_decompress(serialized: bytes) -> Any:
    return pickle_loads(decompress(serialized))


def save_df(df: pd.DataFrame, format: str = "csv") -> bytes:
    pandas_version: int = int(re.sub(r"\D", "", pd.__version__))
    if format == "csv":
        import csv

        csv_buffer = BytesIO()
        if pandas_version >= 120:
            df.to_csv(csv_buffer, index=False, na_rep="_GSK_NA_", escapechar="\\", quoting=csv.QUOTE_NONNUMERIC)
        else:
            csv_buffer.write(
                df.to_csv(index=False, na_rep="_GSK_NA_", escapechar="\\", quoting=csv.QUOTE_NONNUMERIC).encode("utf-8")
            )
            csv_buffer.seek(0)
        return csv_buffer.getvalue()
    else:
        raise ValueError("Invalid method: {method}. Choose 'csv'.")


def compress(data: bytes, method: Optional[str] = "zstd") -> bytes:
    if method == "zstd":
        compressor = ZstdCompressor(level=3, write_checksum=True)
        compressed_data = compressor.compress(data)
    elif method is None:
        compressed_data = data
    # elif compression == "lz4":
    #     import lz4.frame
    #     data = lz4.frame.compress(data, compression_level=3, content_checksum=True)
    else:
        raise ValueError("Invalid compression method: {method}. Choose 'zstd' or None.")

    return compressed_data


def decompress(
    data: bytes, method: Optional[str] = "zstd", max_output_size: int = COMPRESSION_MAX_OUTPUT_SIZE
) -> bytes:
    if method == "zstd":
        decompressor = ZstdDecompressor()
        decompressed_data = decompressor.decompress(data, max_output_size=max_output_size)
    elif method is None:
        decompressed_data = data
    else:
        raise ValueError("Invalid compression method: {method}. Choose 'zstd' or None.")
    return decompressed_data


class GiskardJSONSerializer(json.JSONEncoder):
    def default(self, obj):
        # Wrap numpy bool_ to avoid "TypeError: Object of type bool_ is not JSON serializable"
        if isinstance(obj, np.bool_):
            return super().encode(bool(obj))
        return super().default(obj)



"""
To scan, test and debug your model, you need to wrap it into a Giskard Model.
Your model can use any ML library (``sklearn``, ``catboost``, ``pytorch``, ``tensorflow``,
``huggingface`` and ``langchain``) and can be any Python function that respects the right signature.

You can wrap your model in two different ways:

1. Wrap a prediction function that contains all your data pre-processing steps.
   Prediction function is any Python function that takes input as raw pandas dataframe and returns the probabilities
   for each classification labels (classification) or predictions (regression or text_generation).

   Make sure that:

   - ``prediction_function`` encapsulates all the data pre-processing steps (categorical encoding, numerical scaling, etc.).
   - ``prediction_function(df[feature_names])`` does not return an error message.

2. Wrap a model object in addition to a data pre-processing function.
   Providing the model object to Model allows us to automatically infer the ML library of your model object and
   provide a suitable serialization method (provided by ``save_model`` and ``load_model`` methods).

   This requires:

   - Mandatory: Overriding the model_predict method which should take the input as raw pandas dataframe and return
   the probabilities for each classification labels (classification) or predictions (regression or text_generation).
   - Optional: Our pre-defined serialization and prediction methods cover the ``sklearn``, ``catboost``, ``pytorch``,
   ``tensorflow``, ``huggingface`` and ``langchain`` libraries.
   If none of these libraries are detected, cloudpickle is used as the default for serialization.
   If this fails, we will ask you to also override the ``save_model`` and ``load_model`` methods where you provide your own
   serialization of the model object.

"""

from .automodel import Model

__all__ = ["Model"]



from typing import Hashable

import pandas as pd

from giskard.client.python_utils import warning
from giskard.core.core import SupportedColumnTypes
from giskard.datasets import low_stat_threshold
from giskard.datasets.base import Dataset


def validate_dataset(ds: Dataset):
    validate_dtypes(ds)
    validate_target_exists(ds)
    validate_optional_target(ds)


def validate_optional_target(ds: Dataset):
    if not ds.is_target_given:
        warning(
            "You did not provide the optional argument 'target'. "
            "'target' is the column name in df corresponding to the actual target variable (ground truth)."
        )


def validate_target_exists(ds: Dataset):
    if ds.target is not None and ds.target not in ds.columns:
        raise ValueError(
            "Invalid target parameter:"
            f" '{ds.target}' column is not present in the dataset with columns: {list(ds.columns)}"
        )


def validate_dtypes(ds: Dataset):
    _check_hashability(ds.df)
    _check_mixed_dtypes(ds.df)


def _check_hashability(df):
    """This is a static method that checks if a given pandas DataFrame is hashable or not.

    It checks if all the columns containing object types in the input DataFrame are hashable or not.
    If any column is not hashable, it raises a TypeError indicating which columns are not hashable.

    Parameters
    ----------
    df : pandas.DataFrame
        The DataFrame to be checked for hashability.

    Raises
    ------
    TypeError
        If any column containing object types in the input DataFrame is not hashable.
    """
    if len(df) == 0:
        # Empty df are always hashable
        return

    df_objects = df.select_dtypes(include="object")
    non_hashable_cols = []
    for col in df_objects.columns:
        if not isinstance(df[col].iat[0], Hashable):
            non_hashable_cols.append(col)

    if non_hashable_cols:
        raise TypeError(
            f"The following columns in your df: {non_hashable_cols} are not hashable. "
            f"We currently support only hashable column types such as int, bool, str, tuple and not list or dict."
        )


def _check_mixed_dtypes(df):
    mixed_dtypes = ["mixed", "mixed-integer"]
    mixed_cols = [col for col in df.columns if pd.api.types.infer_dtype(df[col], skipna=True) in mixed_dtypes]

    if len(mixed_cols):
        raise TypeError(
            f"The following columns have mixed data types: {', '.join(mixed_cols)}. "
            "Please make sure that values in each column are of same data type (except NaN)."
        )


def validate_column_types(ds: Dataset):
    """Verifies that declared column_types are correct with regard to SupportedColumnTypes

    Parameters
    ----------
    ds : Dataset
        Dataset to be validated

    """
    if ds.column_types and isinstance(ds.column_types, dict):
        if not set(ds.column_types.values()).issubset(set(column_type.value for column_type in SupportedColumnTypes)):
            raise ValueError(
                f"Invalid column_types parameter: {ds.column_types}"
                + f"Please choose types among {[column_type.value for column_type in SupportedColumnTypes]}."
            )
    else:
        raise ValueError(f"Invalid column_types parameter: {ds.column_types}. Please specify non-empty dictionary.")

    df_columns_set = set(ds.columns)
    df_columns_set.discard(ds.target)
    column_types_set = set(ds.column_types.keys())
    column_types_set.discard(ds.target)

    if column_types_set < df_columns_set:
        missing_columns = df_columns_set - column_types_set
        raise ValueError(
            f"The following keys {list(missing_columns)} are missing from 'column_types'. "
            "Please make sure that the column names in `column_types` covers all the existing "
            "columns in your dataset."
        )


def validate_numeric_columns(ds: Dataset):
    for col, col_type in ds.column_types.items():
        if col == ds.target:
            continue
        if col_type == SupportedColumnTypes.NUMERIC.value:
            try:
                pd.to_numeric(ds.df[col])
            except ValueError:
                warning(
                    f"You declared your column '{col}' as 'numeric' but it contains non-numeric values. "
                    f"Please check if you declared the type of '{col}' correctly in 'column_types'."
                )


def validate_column_categorization(ds: Dataset):
    if len(ds.df) <= low_stat_threshold:
        return

    nuniques = ds.df.nunique()

    for column in ds.columns:
        if column == ds.target:
            continue
        # if a user provided possibly wrong information in column_types or cat_columns about cat columns
        if nuniques[column] <= ds.category_threshold and (
            ds.column_types[column] == SupportedColumnTypes.NUMERIC.value
            or ds.column_types[column] == SupportedColumnTypes.TEXT.value
        ):
            warning(
                f"Feature '{column}' is declared as '{ds.column_types[column]}' but has {nuniques[column]} "
                f"(<= category_threshold={ds.category_threshold}) distinct values. Are "
                "you sure it is not a 'category' feature?"
            )
        # TODO: A bit noisy with a conservative category_threshold, decide on whether to include it or not.
        # if a user provided possibly wrong information in column_types or cat_columns about cat columns
        # elif nuniques[column] > ds.category_threshold and \
        #         ds.column_types[column] == SupportedColumnTypes.CATEGORY.value:
        #     warning(
        #         f"Feature '{column}' is declared as '{ds.column_types[column]}' but has {nuniques[column]} "
        #         f"(> category_threshold={ds.category_threshold}) distinct values. Are "
        #         f"you sure it is a 'category' feature?"
        #     )
        # if a user provided possibly wrong information in column_types about text columns
        else:
            if ds.column_types[column] == SupportedColumnTypes.TEXT.value:
                try:
                    pd.to_numeric(ds.df[column])
                    warning(
                        f"Feature '{column}' is declared as '{ds.column_types[column]}'. Are "
                        "you sure it is not a 'numeric' feature?"
                    )
                except ValueError:
                    pass
            elif ds.column_types[column] == SupportedColumnTypes.NUMERIC.value:
                try:
                    pd.to_numeric(ds.df[column])
                except ValueError:
                    warning(
                        f"Feature '{column}' is declared as '{ds.column_types[column]}'. Are "
                        "you sure it is not a 'text' feature?"
                    )



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from giskard.llm.talk.config import ToolDescription
from giskard.llm.talk.tools.base import BaseTool

if TYPE_CHECKING:
    from giskard.scanner.report import ScanReport


class IssuesScannerTool(BaseTool):
    """Issues Scanner Tool.

    Attributes
    ----------
    default_name : str
        The default name of the Tool. Can be re-defined with constructor.
    default_description: str
        The default description of the Tool's functioning. Can be re-defined with constructor.
    """

    default_name: str = "issues_scanner"
    default_description: str = ToolDescription.ISSUES_SCANNER.value
    _output_template: str = "ML model performance issues scanner result:\n {scan_report}"

    def __init__(self, scan_report: ScanReport, name: Optional[str] = None, description: Optional[str] = None):
        """Constructor of the class.

        Parameters
        ----------
        scan_report : ScanReport
            The Giskard ScanReport object.
        name : str, optional
            The name of the Tool.
            If not set, the `default_name` is used.
        description : str, optional
            The description of the Tool.
            If not set, the `default_description` is used.
        """
        super().__init__(name, description)
        self._scan_report = scan_report

    @property
    def specification(self) -> str:
        """Return the Tool's specification in a JSON Schema format.

        Returns
        -------
        str
            The Tool's specification.
        """
        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": self.description,
            },
        }

    def __call__(self) -> str:
        """Execute the Tool's functionality.

        Return the Giskard Scan result.

        Returns
        -------
        str
            The Giskard Scan's result.

        Raises
        ------
        AttributeError
            If the 'scan_report' was not provided.
        """
        if not self._scan_report:
            raise AttributeError(
                "To obtain information about issues detected by the Giskard Scan, provide the 'scan_report' argument."
            )
        scan_report = self._scan_report.to_markdown(template="huggingface")
        return self._output_template.format(scan_report=scan_report)



from typing import Dict, List, Optional, Sequence
from typing_extensions import deprecated

import json
from abc import ABC, abstractmethod

from ..config import LLMConfigurationError
from ..errors import LLMImportError
from . import LLMClient
from .base import ChatMessage

try:
    import boto3  # noqa: F401
except ImportError as err:
    raise LLMImportError(
        flavor="llm", msg="To use Bedrock models, please install the `boto3` package with `pip install boto3`"
    ) from err


class BaseBedrockClient(LLMClient, ABC):
    def __init__(self, bedrock_runtime_client, model: str):
        self._client = bedrock_runtime_client
        self.model = model

    @abstractmethod
    def _format_body(
        self,
        messages: Sequence[ChatMessage],
        temperature: float = 1,
        max_tokens: Optional[int] = 1000,
        caller_id: Optional[str] = None,
        seed: Optional[int] = None,
        format=None,
    ) -> Dict:
        ...

    @abstractmethod
    def _parse_completion(self, completion, caller_id: Optional[str] = None) -> ChatMessage:
        ...

    def complete(
        self,
        messages: Sequence[ChatMessage],
        temperature: float = 1,
        max_tokens: Optional[int] = 1000,
        caller_id: Optional[str] = None,
        seed: Optional[int] = None,
        format=None,
    ) -> ChatMessage:
        # create the json body to send to the API
        body = self._format_body(messages, temperature, max_tokens, caller_id, seed, format)

        # invoke the model and get the response
        try:
            accept = "application/json"
            contentType = "application/json"
            response = self._client.invoke_model(body=body, modelId=self.model, accept=accept, contentType=contentType)
            completion = json.loads(response.get("body").read())
        except RuntimeError as err:
            raise LLMConfigurationError("Could not get response from Bedrock API") from err

        return self._parse_completion(completion, caller_id)

    def get_config(self) -> dict:
        """Return the configuration of the LLM client."""
        return {"client_type": self.__class__.__name__, "model": self.model}


@deprecated("ClaudeBedrockClient is deprecated: https://docs.giskard.ai/en/latest/open_source/setting_up/index.html")
class ClaudeBedrockClient(BaseBedrockClient):
    def __init__(
        self,
        bedrock_runtime_client,
        model: str = "anthropic.claude-3-sonnet-20240229-v1:0",
        anthropic_version: str = "bedrock-2023-05-31",
    ):
        # only supporting claude 3
        if "claude-3" not in model:
            raise LLMConfigurationError(f"Only claude-3 models are supported as of now, got {self.model}")

        super().__init__(bedrock_runtime_client, model)
        self.anthropic_version = anthropic_version

    def _format_body(
        self,
        messages: Sequence[ChatMessage],
        temperature: float = 1,
        max_tokens: Optional[int] = 1000,
        caller_id: Optional[str] = None,
        seed: Optional[int] = None,
        format=None,
    ) -> Dict:
        input_msg_prompt: List = []
        system_prompts = []

        for msg in messages:
            # System prompt is a specific parameter in Claude
            if msg.role.lower() == "system":
                system_prompts.append(msg.content)
                continue

            # Only role user and assistant are allowed
            role = msg.role.lower()
            role = role if role in ["assistant", "user"] else "user"

            # Consecutive messages need to be grouped
            last_message = None if len(input_msg_prompt) == 0 else input_msg_prompt[-1]
            if last_message is not None and last_message["role"] == role:
                last_message["content"].append({"type": "text", "text": msg.content})
                continue

            input_msg_prompt.append({"role": role, "content": [{"type": "text", "text": msg.content}]})

        return json.dumps(
            {
                "anthropic_version": "bedrock-2023-05-31",
                "max_tokens": max_tokens,
                "temperature": temperature,
                "system": "\n".join(system_prompts),
                "messages": input_msg_prompt,
            }
        )

    def _parse_completion(self, completion, caller_id: Optional[str] = None) -> ChatMessage:
        self.logger.log_call(
            prompt_tokens=completion["usage"]["input_tokens"],
            sampled_tokens=completion["usage"]["output_tokens"],
            model=self.model,
            client_class=self.__class__.__name__,
            caller_id=caller_id,
        )

        msg = completion["content"][0]["text"]
        return ChatMessage(role="assistant", content=msg)


@deprecated("LLamaBedrockClient is deprecated: https://docs.giskard.ai/en/latest/open_source/setting_up/index.html")
class LLamaBedrockClient(BaseBedrockClient):
    def __init__(self, bedrock_runtime_client, model: str = "meta.llama3-8b-instruct-v1:0"):
        # only supporting llama
        if "llama" not in model:
            raise LLMConfigurationError(f"Only Llama models are supported as of now, got {self.model}")

        super().__init__(bedrock_runtime_client, model)

    def _format_body(
        self,
        messages: Sequence[ChatMessage],
        temperature: float = 1,
        max_tokens: Optional[int] = 1000,
        caller_id: Optional[str] = None,
        seed: Optional[int] = None,
        format=None,
    ) -> Dict:
        # Create the messages format needed for llama bedrock specifically
        prompts = []
        for msg in messages:
            prompts.append(f"# {msg.role}:\n{msg.content}\n")

        # create the json body to send to the API
        messages = "\n".join(prompts)
        return json.dumps(
            {
                "max_gen_len": max_tokens,
                "temperature": temperature,
                "prompt": f"{messages}\n# assistant:\n",
            }
        )

    def _parse_completion(self, completion, caller_id: Optional[str] = None) -> ChatMessage:
        self.logger.log_call(
            prompt_tokens=completion["prompt_token_count"],
            sampled_tokens=completion["generation_token_count"],
            model=self.model,
            client_class=self.__class__.__name__,
            caller_id=caller_id,
        )

        msg = completion["generation"]
        return ChatMessage(role="assistant", content=msg)



from typing import List

import itertools
import random
from collections.abc import Iterator
from functools import lru_cache

import numpy as np


def map_to_tuples(data: Iterator):
    item = next(data)
    data = itertools.chain([item], data)

    if isinstance(item, tuple):  # nothing to do
        return data

    return map(lambda x: (x,), data)


def np_types_to_native(some_list: List):
    if some_list is not None:
        return [np_type_to_native(i) for i in some_list]
    return some_list


def np_type_to_native(i):
    return i.item() if isinstance(i, np.generic) else i


def fix_seed(seed=1337):
    random.seed(seed)
    np.random.seed(seed)
    try:
        import torch

        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
    except ImportError:
        pass


@lru_cache(None)
def warn_once(logger, msg: str):
    logger.warning(msg)



from typing import Sequence

from ..decorators import detector
from ..issues import Ethical
from .base_detector import BaseTextPerturbationDetector
from .text_transformations import TextTransformation


@detector(
    name="ethical_bias",
    tags=["ethical_bias", "robustness", "classification", "regression"],
)
class EthicalBiasDetector(BaseTextPerturbationDetector):
    """Detects ethical bias in a model by applying text perturbations to the input data.

    You can explicitly run this detector by adding the tag "ethical_bias" in the `only` parameter of the scan method.

    By default, we perform specific metamorphic testing aimed at detecting bias in the model predictions based on
    transformation of gender, nationality, or religious terms in the textual features.

    As an example, for a sentiment analysis model we will transform a sentence like "She is such a talented singer"
    into "He is such a talented singer" and check if the model prediction changes. If it does systematically, it
    means that the model has some form of gender bias.
    """

    _issue_group = Ethical
    _taxonomy = ["avid-effect:ethics:E0101", "avid-effect:performance:P0201"]

    def _get_default_transformations(self) -> Sequence[TextTransformation]:
        from .text_transformations import (
            TextGenderTransformation,
            TextNationalityTransformation,
            TextReligionTransformation,
        )

        return [
            TextGenderTransformation,
            TextReligionTransformation,
            TextNationalityTransformation,
        ]



import inspect
import re

from .registry import DetectorRegistry


def detector(name=None, tags=None):
    if inspect.isclass(name):
        # If the decorator is used without arguments, the first argument is the class
        cls = name
        DetectorRegistry.register(_to_snake_case(cls.__name__), cls)
        return cls

    def inner(cls):
        DetectorRegistry.register(name or _to_snake_case(cls.__name__), cls, tags)
        return cls

    return inner


def _to_snake_case(string) -> str:
    s = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", string)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s).lower()



from typing import Any, Callable, Iterable, Optional, Tuple, Union

import logging
import pickle
from abc import ABC, abstractmethod
from inspect import isfunction, signature
from pathlib import Path

import cloudpickle
import numpy as np
import pandas as pd
import yaml

from giskard.exceptions.giskard_exception import python_env_exception_helper

from ...core.core import ModelType
from ...core.validation import configured_validate_arguments
from ..utils import warn_once
from .model import BaseModel

logger = logging.getLogger(__name__)


class WrapperModel(BaseModel, ABC):
    """Base class for model wrappers.

    This is subclass of a :class:`BaseModel` that wraps an existing model
    object and uses it to make inference.

    This class introduces a `data_preprocessing_function` which can be used
    to preprocess incoming data before it is passed to the underlying model.
    """

    @configured_validate_arguments
    def __init__(
        self,
        model: Any,
        model_type: ModelType,
        data_preprocessing_function: Optional[Callable[[pd.DataFrame], Any]] = None,
        model_postprocessing_function: Optional[Callable[[Any], Any]] = None,
        name: Optional[str] = None,
        feature_names: Optional[Iterable] = None,
        classification_threshold: Optional[float] = 0.5,
        classification_labels: Optional[Iterable] = None,
        id: Optional[str] = None,
        batch_size: Optional[int] = None,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        model : Any
            The model that will be wrapped.
        model_type : ModelType
            The type of the model. Must be a value from the :class:`ModelType`
            enumeration.
        data_preprocessing_function : Optional[Callable[[pd.DataFrame], Any]]
            A function that will be applied to incoming data. Default is ``None``.
        model_postprocessing_function : Optional[Callable[[Any], Any]]
            A function that will be applied to the model's predictions. Default
            is ``None``.
        name : Optional[str]
            A name for the wrapper. Default is ``None``.
        feature_names : Optional[Iterable]
            A list of feature names. Default is ``None``.
        classification_threshold : Optional[float]
            The probability threshold for classification. Default is 0.5.
        classification_labels : Optional[Iterable]
            A list of classification labels. Default is None.
        batch_size : Optional[int]
            The batch size to use for inference. Default is ``None``, which
            means inference will be done on the full dataframe.
        """
        super().__init__(
            model_type=model_type,
            name=name,
            feature_names=feature_names,
            classification_threshold=classification_threshold,
            classification_labels=classification_labels,
            id=id,
            batch_size=batch_size,
            **kwargs,
        )
        self.model = model
        self.data_preprocessing_function = data_preprocessing_function
        self.model_postprocessing_function = model_postprocessing_function
        self.batch_size = batch_size

        # TODO: refactor this into validate_args or another decorator @validate_sign
        if self.data_preprocessing_function and isfunction(self.data_preprocessing_function):
            sign_len = len(signature(self.data_preprocessing_function).parameters)
            if sign_len != 1:
                raise ValueError(
                    f"data_preprocessing_function only takes 1 argument (a pandas.DataFrame) but {sign_len} were provided."
                )
        if self.model_postprocessing_function:
            sign_len = len(signature(self.model_postprocessing_function).parameters)
            if sign_len != 1:
                raise ValueError(f"model_postprocessing_function only takes 1 argument but {sign_len} were provided.")

    def _preprocess(self, data):
        if self.data_preprocessing_function:
            return self.data_preprocessing_function(data)
        return data

    def _postprocess(self, raw_predictions):
        # User specified a custom postprocessing function
        if self.model_postprocessing_function:
            raw_predictions = self.model_postprocessing_function(raw_predictions)

        # Convert predictions to numpy array
        raw_predictions = self._convert_to_numpy(raw_predictions)

        # We try to automatically fix issues in the output shape
        raw_predictions = self._possibly_fix_predictions_shape(raw_predictions)

        return raw_predictions

    def _convert_to_numpy(self, raw_predictions):
        return np.asarray(raw_predictions)

    @configured_validate_arguments
    def predict_df(self, df: pd.DataFrame, *_args, **_kwargs):
        if self.batch_size and self.batch_size > 0:
            dfs = np.array_split(df, np.arange(self.batch_size, len(df), self.batch_size))
        else:
            dfs = [df]

        outputs = []
        for batch in map(self._preprocess, dfs):
            output = self.model_predict(batch)
            output = self._postprocess(output)
            outputs.append(output)

        raw_prediction = np.concatenate(outputs)

        return raw_prediction

    def _possibly_fix_predictions_shape(self, raw_predictions: np.ndarray):
        if not self.is_classification:
            return raw_predictions

        # Ensure this is 2-dimensional
        if raw_predictions.ndim <= 1:
            raw_predictions = raw_predictions.reshape(-1, 1)

        # Fix possible extra dimensions (e.g. batch dimension which was not squeezed)
        if raw_predictions.ndim > 2:
            warn_once(
                logger,
                f"\nThe output of your model has shape {raw_predictions.shape}, but we expect a shape (n_entries, n_classes). \n"
                "We will attempt to automatically reshape the output to match this format, please check that the results are consistent.",
            )

            raw_predictions = raw_predictions.squeeze(tuple(range(1, raw_predictions.ndim - 1)))

            if raw_predictions.ndim > 2:
                raise ValueError(
                    f"The output of your model has shape {raw_predictions.shape}, but we expect it to be (n_entries, n_classes)."
                )

        # E.g. for binary classification, prediction should be of the form `(p, 1 - p)`.
        # If a binary classifier returns a single prediction `p`, we try to infer the second class
        # prediction as `1 - p`.
        if self.is_binary_classification and raw_predictions.shape[-1] == 1:
            warn_once(
                logger,
                "Please make sure that your model's output corresponds "
                "to the second label in classification_labels.",
            )

            raw_predictions = np.append(1 - raw_predictions, raw_predictions, axis=1)

        # For classification models, the last dimension must be equal to the number of classes
        if raw_predictions.shape[-1] != len(self.classification_labels):
            raise ValueError(
                f"The output of your model has shape {raw_predictions.shape}, but we expect it to be (n_entries, n_classes), \n"
                f"where `n_classes` is the number of classes in your model output ({len(self.classification_labels)} in this case)."
            )

        return raw_predictions

    @abstractmethod
    def model_predict(self, data):
        """Performs the model inference/forward pass.

        Parameters
        ----------
        data : Any
            The input data for making predictions. If you did not specify a
            `data_preprocessing_function`, this will be a :class:`pd.DataFrame`,
            otherwise it will be whatever the `data_preprocessing_function`
            returns.

        Returns
        -------
        numpy.ndarray
            If the model is ``classification``, it should return an array of
            probabilities of shape ``(num_entries, num_classes)``.
            If the model is ``regression`` or ``text_generation``, it should
            return an array of ``num_entries`` predictions.
        """
        ...

    def save(self, local_path: Union[str, Path], *args, **kwargs) -> None:
        super().save(local_path, *args, **kwargs)
        self.save_wrapper_meta(local_path)
        if self.data_preprocessing_function:
            self.save_data_preprocessing_function(local_path)
        if self.model_postprocessing_function:
            self.save_model_postprocessing_function(local_path)

    @abstractmethod
    def save_model(self, path: Union[str, Path], *args, **kwargs) -> None:
        """Saves the wrapped ``model`` object.

        Parameters
        ----------
        path : Union[str, Path]
            Path to which the model should be saved.
        """
        ...

    def save_data_preprocessing_function(self, local_path: Union[str, Path], *_args, **_kwargs):
        with open(Path(local_path) / "giskard-data-preprocessing-function.pkl", "wb") as f:
            cloudpickle.dump(self.data_preprocessing_function, f, protocol=pickle.DEFAULT_PROTOCOL)

    def save_model_postprocessing_function(self, local_path: Union[str, Path], *_args, **_kwargs):
        with open(Path(local_path) / "giskard-model-postprocessing-function.pkl", "wb") as f:
            cloudpickle.dump(self.model_postprocessing_function, f, protocol=pickle.DEFAULT_PROTOCOL)

    def save_wrapper_meta(self, local_path, *_args, **_kwargs):
        with open(Path(local_path) / "giskard-model-wrapper-meta.yaml", "w", encoding="utf-8") as f:
            yaml.dump(
                {
                    "batch_size": self.batch_size,
                },
                f,
                default_flow_style=False,
            )

    @classmethod
    def load(cls, local_dir, model_py_ver: Optional[Tuple[str, str, str]] = None, *args, **kwargs):
        constructor_params = cls.load_constructor_params(local_dir, *args, **kwargs)

        if model_py_ver is None:
            # Try to extract Python version from meta info under local dir
            meta_response, _ = cls.read_meta_from_local_dir(local_dir)
            model_py_ver = (
                tuple(meta_response.languageVersion.split(".")) if "PYTHON" == meta_response.language.upper() else None
            )

        return cls(model=cls.load_model(local_dir, model_py_ver), **constructor_params)

    @classmethod
    def load_constructor_params(cls, local_dir, model_py_ver: Optional[Tuple[str, str, str]] = None, *_args, **kwargs):
        params = cls.load_wrapper_meta(local_dir)
        params["data_preprocessing_function"] = cls.load_data_preprocessing_function(local_dir)
        params["model_postprocessing_function"] = cls.load_model_postprocessing_function(local_dir)
        params.update(kwargs)

        extra_meta, meta = cls.read_meta_from_local_dir(local_dir)
        constructor_params = meta.__dict__
        constructor_params["id"] = extra_meta.id
        constructor_params = constructor_params.copy()
        constructor_params.update(params)

        return constructor_params

    @classmethod
    @abstractmethod
    def load_model(cls, path: Union[str, Path], model_py_ver: Optional[Tuple[str, str, str]] = None, *args, **kwargs):
        """Loads the wrapped ``model`` object.

        Parameters
        ----------
        path : Union[str, Path]
            Path from which the model should be loaded.
        model_py_ver : Optional[Tuple[str, str, str]]
            Python version used to save the model, to validate if model loading failed.
        """
        ...

    @classmethod
    def load_data_preprocessing_function(cls, local_path: Union[str, Path], *_args, **_kwargs):
        local_path = Path(local_path)
        file_path = local_path / "giskard-data-preprocessing-function.pkl"
        if file_path.exists():
            with open(file_path, "rb") as f:
                try:
                    # According to https://github.com/cloudpipe/cloudpickle#cloudpickle:
                    # Cloudpickle can only be used to send objects between the exact same version of Python.
                    return cloudpickle.load(f)
                except Exception as e:
                    raise python_env_exception_helper("Data Preprocessing Function", e)
        return None

    @classmethod
    def load_model_postprocessing_function(cls, local_path: Union[str, Path], *_args, **_kwargs):
        local_path = Path(local_path)
        file_path = local_path / "giskard-model-postprocessing-function.pkl"
        if file_path.exists():
            with open(file_path, "rb") as f:
                try:
                    # According to https://github.com/cloudpipe/cloudpickle#cloudpickle:
                    # Cloudpickle can only be used to send objects between the exact same version of Python.
                    return cloudpickle.load(f)
                except Exception as e:
                    raise python_env_exception_helper("Data Postprocessing Function", e)
        return None

    @classmethod
    def load_wrapper_meta(cls, local_dir, *args, **kwargs):
        wrapper_meta_file = Path(local_dir) / "giskard-model-wrapper-meta.yaml"
        if wrapper_meta_file.exists():
            with open(wrapper_meta_file, encoding="utf-8") as f:
                wrapper_meta = yaml.load(f, Loader=yaml.Loader)
                wrapper_meta["batch_size"] = int(wrapper_meta["batch_size"]) if wrapper_meta["batch_size"] else None
                return wrapper_meta
        else:
            # ensuring backward compatibility
            return {"batch_size": None}

    def to_mlflow(self, artifact_path: str = "prediction-function-from-giskard", *_args, **_kwargs):
        import mlflow

        def _giskard_predict(df):
            return self.predict(df)

        class MLflowModel(mlflow.pyfunc.PythonModel):
            def predict(self, df):
                return _giskard_predict(df)

        mlflow_model = MLflowModel()
        return mlflow.pyfunc.log_model(artifact_path=artifact_path, python_model=mlflow_model)



from ..decorators import detector
from ..issues import IssueLevel, SensitiveInformationDisclosure
from .base import RequirementBasedDetector

SENSITIVE_INFORMATION_DISCLOSURE_ISSUE_DESCRIPTION = (
    "Disclosure of sensitive information, for example the agent should not divulge confidential information such as "
    "PIIs, medical records, passwords, credentials or other secrets."
)


@detector("llm_information_disclosure", tags=["information_disclosure", "llm", "generative", "text_generation"])
class LLMInformationDisclosureDetector(RequirementBasedDetector):
    """Detects sensitive information disclosure in LLM-based models.

    You can explicitly run this detector by adding the tag "information_disclosure" in the `only` parameter of the scan method.

    This detector checks that the model does divulge or hallucinate sensitive or confidential information in its
    responses. We probe the model by generating ad hoc adversarial inputs and checking that the model outputs do not
    contain information that could be considered sensitive, such as personal identifiable information (PII) or secret
    credentials.

    In some cases, this can produce false positives if the model is supposed to return sensitive information (e.g.
    contact information for a business). We still recommend to carefully review the detections, as they may reveal
    undesired availability of private information to the model (for example, confidential data acquired during fine
    tuning), or the tendency to hallucinate information such as phone numbers or personal emails even if those details
    were not provided to the model.
    """

    _issue_group = SensitiveInformationDisclosure
    _issue_level = IssueLevel.MEDIUM
    _taxonomy = ["avid-effect:performance:P0301"]

    def get_issue_description(self) -> str:
        return SENSITIVE_INFORMATION_DISCLOSURE_ISSUE_DESCRIPTION



import typing
from typing import List, Optional

import inspect
import numbers
import re
import uuid
from collections import Counter

import numpy as np
import pandas as pd
from scipy.stats import chi2, ks_2samp
from scipy.stats.stats import Ks_2sampResult, wasserstein_distance

from giskard.core.test_result import TestMessage, TestMessageLevel, TestResult
from giskard.datasets.base import Dataset
from giskard.models.base import BaseModel
from giskard.registry.decorators import test
from giskard.registry.slicing_function import SlicingFunction
from giskard.testing.utils.utils import check_slice_not_empty, validate_classification_label

from . import debug_description_prefix

other_modalities_pattern = "^other_modalities_[a-z0-9]{32}$"


def check_if_debuggable(actual_ds, reference_ds):
    if actual_ds.df.empty:
        raise ValueError("Your actual_dataset is empty. Debug is not defined for this case.")
    if reference_ds.df.empty:
        raise ValueError("Your reference_dataset is empty. Debug is not defined for this case.")
    if actual_ds.id == reference_ds.id:
        raise ValueError(
            "You passed the same dataset as actual_dataset and reference_dataset. "
            "Debug is not defined for this case."
        )
    if actual_ds.df.equals(reference_ds.df):
        raise ValueError(
            "Your actual_dataset is identical to your reference_dataset. " "Debug is not defined for this case."
        )


def _calculate_psi(category, actual_distribution, expected_distribution):
    # To use log and avoid zero distribution probability,
    # we bound distribution probability by min_distribution_probability
    min_distribution_probability = 0.0001

    expected_distribution_bounded = max(expected_distribution[category], min_distribution_probability)
    actual_distribution_bounded = max(actual_distribution[category], min_distribution_probability)
    modality_psi = (expected_distribution_bounded - actual_distribution_bounded) * np.log(
        expected_distribution_bounded / actual_distribution_bounded
    )
    return modality_psi


def _calculate_frequencies(actual_series, reference_series, max_categories=None):
    all_modalities = list(set(reference_series).union(set(actual_series)))
    if max_categories is not None and len(all_modalities) > max_categories:
        var_count_expected = dict(Counter(reference_series).most_common(max_categories))
        other_modalities_key = "other_modalities_" + uuid.uuid1().hex
        var_count_expected[other_modalities_key] = len(reference_series) - sum(var_count_expected.values())
        categories_list = list(var_count_expected.keys())

        var_count_actual = Counter(actual_series)
        # For test data, we take the same category names as expected_data
        var_count_actual = {i: var_count_actual[i] for i in categories_list}
        var_count_actual[other_modalities_key] = len(actual_series) - sum(var_count_actual.values())

        all_modalities = categories_list
    else:
        var_count_expected = Counter(reference_series)
        var_count_actual = Counter(actual_series)
    expected_frequencies = np.array([var_count_expected[i] for i in all_modalities])
    actual_frequencies = np.array([var_count_actual[i] for i in all_modalities])
    return all_modalities, actual_frequencies, expected_frequencies


def _calculate_drift_psi(actual_series, reference_series, max_categories):
    (
        all_modalities,
        actual_frequencies,
        expected_frequencies,
    ) = _calculate_frequencies(actual_series, reference_series, max_categories)
    expected_distribution = expected_frequencies / len(reference_series)
    actual_distribution = actual_frequencies / len(actual_series)
    total_psi = 0
    output_data = []
    for category, modality in enumerate(all_modalities):
        modality_psi = _calculate_psi(category, actual_distribution, expected_distribution)
        total_psi += modality_psi
        output_data.append(
            {
                "Modality": modality,
                "Reference_distribution": expected_distribution[category],
                "Actual_distribution": expected_distribution[category],
                "Psi": modality_psi,
            }
        )
    return total_psi, pd.DataFrame(output_data)


def _calculate_ks(actual_series, reference_series) -> Ks_2sampResult:
    return ks_2samp(reference_series, actual_series)


def _calculate_earth_movers_distance(actual_series, reference_series):
    unique_reference = np.unique(reference_series)
    unique_actual = np.unique(actual_series)
    sample_space = list(set(unique_reference).union(set(unique_actual)))
    val_max = max(sample_space)
    val_min = min(sample_space)
    if val_max == val_min:
        metric = 0
    else:
        # Normalizing reference_series and actual_series for comparison purposes
        reference_series = (reference_series - val_min) / (val_max - val_min)
        actual_series = (actual_series - val_min) / (val_max - val_min)

        metric = wasserstein_distance(reference_series, actual_series)
    return metric


def _calculate_chi_square(actual_series, reference_series, max_categories):
    (
        all_modalities,
        actual_frequencies,
        expected_frequencies,
    ) = _calculate_frequencies(actual_series, reference_series, max_categories)
    chi_square = 0
    # it's necessary for comparison purposes to normalize expected_frequencies
    # so that reference and actual has the same size
    # See https://github.com/scipy/scipy/blob/v1.8.0/scipy/stats/_stats_py.py#L6787
    k_norm = actual_series.shape[0] / reference_series.shape[0]
    output_data = []
    for i, modality in enumerate(all_modalities):
        chi_square_value = (actual_frequencies[i] - expected_frequencies[i] * k_norm) ** 2 / (
            expected_frequencies[i] * k_norm
        )
        chi_square += chi_square_value

        output_data.append(
            {
                "Modality": modality,
                "Reference_frequencies": expected_frequencies[i],
                "Actual_frequencies": actual_frequencies[i],
                "Chi_square": chi_square_value,
            }
        )

    # if reference_series and actual_series has only one modality it turns nan (len(all_modalities)=1)
    if len(all_modalities) > 1:
        chi_cdf = chi2.cdf(chi_square, len(all_modalities) - 1)
        p_value = 1 - chi_cdf if chi_cdf != 0 else 0
    else:
        p_value = 0
    return chi_square, p_value, pd.DataFrame(output_data)


def _validate_feature_type(gsk_dataset, column_name, feature_type):
    assert (
        gsk_dataset.column_types[column_name] == feature_type
    ), f'Column "{column_name}" is not of type "{feature_type}"'


def _validate_column_name(actual_ds, reference_ds, column_name):
    assert (
        column_name in actual_ds.columns
    ), f'"{column_name}" is not a column of Actual Dataset Columns: {", ".join(actual_ds.columns)}'
    assert (
        column_name in reference_ds.columns
    ), f'"{column_name}" is not a column of Reference Dataset Columns: {", ".join(reference_ds.columns)}'


def _validate_series_notempty(actual_series, reference_series):
    if actual_series.empty:
        raise ValueError("Actual Series computed from the column is empty")
    if reference_series.empty:
        raise ValueError("Reference Series computed from the column is empty")


def _extract_series(actual_ds, reference_ds, column_name, feature_type):
    _validate_column_name(actual_ds, reference_ds, column_name)
    _validate_feature_type(actual_ds, column_name, feature_type)
    _validate_feature_type(reference_ds, column_name, feature_type)
    actual_series = actual_ds.df[column_name]
    reference_series = reference_ds.df[column_name]
    _validate_series_notempty(actual_series, reference_series)
    return actual_series, reference_series


@test(
    name="Categorical drift (PSI)",
    debug_description=debug_description_prefix + "with <b>the categories that have drifted "
    "the most from the 'actual_dataset'</b>.",
)
def test_drift_psi(
    actual_dataset: Dataset,
    reference_dataset: Dataset,
    column_name: str,
    slicing_function: Optional[SlicingFunction] = None,
    threshold: float = 0.2,
    max_categories: int = 20,
    psi_contribution_percent: float = 0.2,
) -> TestResult:
    """Test categorical drift by PSI score.

    It tests that the PSI score between the actual and reference datasets is
    below the threshold for a given categorical feature.

    For example, the test is passed when the PSI score of gender between
    reference and actual sets is below threshold 0.2.

    Parameters
    ----------
    actual_dataset : Dataset
        Actual dataset used to compute the test
    reference_dataset : Dataset
        Reference dataset used to compute the test
    column_name : str
        Name of column with categorical feature
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on both actual and reference datasets.
    threshold : float
        Threshold value for PSI. Default value = 0.2.
    max_categories : int
        The maximum categories to compute the PSI score.
    psi_contribution_percent : float
        The ratio between the PSI score of a given category over the total PSI
        score of the categorical variable. If there is a drift, the test
        provides all the categories that have a PSI contribution over than this
        ratio.
    debug : bool
        If True and the test fails, a dataset will be provided containing the
        actual_dataset rows with the categories that have drifted the most (more
        than psi_contribution_percent of the total PSI score). Default is False.

    Returns
    -------
    TestResult
        The test result.
    """
    if slicing_function:
        test_name = "test_drift_psi"
        actual_dataset = actual_dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=actual_dataset, dataset_name="actual_dataset", test_name=test_name)
        reference_dataset = reference_dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=reference_dataset, dataset_name="reference_dataset", test_name=test_name)

    actual_series, reference_series = _extract_series(actual_dataset, reference_dataset, column_name, "category")

    messages, passed, total_psi, output_data = _test_series_drift_psi(
        actual_series,
        reference_series,
        "data",
        max_categories,
        psi_contribution_percent,
        threshold,
    )

    # --- debug ---
    output_ds = list()
    if not passed:
        check_if_debuggable(actual_dataset, reference_dataset)
        main_drifting_modalities_bool = output_data["Psi"] > psi_contribution_percent * total_psi
        modalities_list = output_data[main_drifting_modalities_bool]["Modality"].tolist()
        if modalities_list:
            filtered_modalities = [w for w in modalities_list if not re.match(other_modalities_pattern, str(w))]
            output_ds.append(
                actual_dataset.slice(lambda df: df.loc[actual_series.isin(filtered_modalities)], row_level=False)
            )
            test_name = inspect.stack()[0][3]
            if len(output_ds[0].df.index) == 0:
                raise ValueError(
                    test_name
                    + f": the categories {filtered_modalities} completely drifted as they are not present in the "
                    f"'actual_dataset'"
                )
    # ---

    return TestResult(
        actual_slices_size=[len(actual_series)],
        reference_slices_size=[len(reference_series)],
        passed=passed,
        metric=total_psi,
        messages=messages,
        output_ds=output_ds,
    )


@test(
    name="Categorical drift (Chi-squared)",
    debug_description=debug_description_prefix + "with <b>the categories that have drifted "
    "the most from the 'actual_dataset'</b>.",
)
def test_drift_chi_square(
    actual_dataset: Dataset,
    reference_dataset: Dataset,
    column_name: str,
    slicing_function: Optional[SlicingFunction] = None,
    threshold: float = 0.05,
    max_categories: int = 20,
    chi_square_contribution_percent: float = 0.2,
) -> TestResult:
    """Tests drift by chi-squared.

    The test checks if the p-value of the chi-square test between the actual and
    reference datasets is above a threshold for a given categorical feature.

    Parameters
    ----------
    actual_dataset : Dataset
        Actual dataset used to compute the test
    reference_dataset : Dataset
        Reference dataset used to compute the test
    column_name : str
        Name of column with categorical feature
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on both actual and reference datasets
    threshold : float
        Threshold for p-value of chi-square. Default is 0.05.
    max_categories : int
        The maximum categories to compute the chi-square.
    chi_square_contribution_percent : float
        the ratio between the Chi-Square value of a given category over the
        total chi-square value of the categorical variable. If there is a drift,
        the test provides all the categories that have a PSI contribution over
        than this ratio.
    debug : bool
        If True and the test fails, a dataset will be provided containing the
        actual_dataset rows with the categories that have drifted the most
        (more than chi_square_contribution_percent of the total chi squared
        score).

    Returns
    -------
    TestResult
        The test result.
    """
    if slicing_function:
        test_name = "test_drift_chi_square"
        actual_dataset = actual_dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=actual_dataset, dataset_name="actual_dataset", test_name=test_name)
        reference_dataset = reference_dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=reference_dataset, dataset_name="reference_dataset", test_name=test_name)

    actual_series, reference_series = _extract_series(actual_dataset, reference_dataset, column_name, "category")

    messages, chi_square, p_value, passed, output_data = _test_series_drift_chi(
        actual_series,
        reference_series,
        "data",
        chi_square_contribution_percent,
        max_categories,
        threshold,
    )

    # --- debug ---
    output_ds = list()
    if not passed:
        check_if_debuggable(actual_dataset, reference_dataset)
        main_drifting_modalities_bool = output_data["Chi_square"] > chi_square_contribution_percent * chi_square
        modalities_list = output_data[main_drifting_modalities_bool]["Modality"].tolist()
        if modalities_list:
            filtered_modalities = [w for w in modalities_list if not re.match(other_modalities_pattern, str(w))]
            output_ds.append(
                actual_dataset.slice(lambda df: df.loc[actual_series.isin(filtered_modalities)], row_level=False)
            )
    # ---

    return TestResult(
        actual_slices_size=[len(actual_series)],
        reference_slices_size=[len(reference_series)],
        passed=passed,
        metric=p_value,
        messages=messages,
        output_ds=output_ds,
    )


@test(name="Numerical drift (Kolmogorov-Smirnov)")
def test_drift_ks(
    actual_dataset: Dataset,
    reference_dataset: Dataset,
    column_name: str,
    slicing_function: Optional[SlicingFunction] = None,
    threshold: float = 0.05,
) -> TestResult:
    """Test drift with a Kolmogorov-Smirnov test.

    Test if the p-value of the Kolmogorov-Smirnov test between the actual and
    reference datasets is above a threshold for a given numerical feature.

    For example, if the threshold is set to 0.05, the test is passed when the
    p-value of the KS test of the numerical variable between the actual and
    reference datasets is higher than 0.05. It means that the null hypothesis
    (no drift) cannot be rejected at 5% confidence level and that we cannot
    assume drift for this variable.

    Parameters
    ----------
    actual_dataset : Dataset
        Actual dataset used to compute the test.
    reference_dataset : Dataset
        Reference dataset used to compute the test.
    column_name : str
        Name of column with numerical feature.
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on both actual and reference datasets.
    threshold : float
        Threshold for p-value of Kolmogorov-Smirnov test. Default is = 0.05.

    Returns
    -------
    TestResult
        The test result.
    """
    if slicing_function:
        test_name = "test_drift_ks"
        actual_dataset = actual_dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=actual_dataset, dataset_name="actual_dataset", test_name=test_name)
        reference_dataset = reference_dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=reference_dataset, dataset_name="reference_dataset", test_name=test_name)

    actual_series, reference_series = _extract_series(actual_dataset, reference_dataset, column_name, "numeric")

    result = _calculate_ks(actual_series, reference_series)

    passed = bool(result.pvalue >= threshold)

    messages = _generate_message_ks(passed, result, threshold, "data")

    return TestResult(
        actual_slices_size=[len(actual_series)],
        reference_slices_size=[len(reference_series)],
        passed=passed,
        metric=result.pvalue,
        messages=messages,
    )


@test(name="Numerical drift (Earth mover's distance)")
def test_drift_earth_movers_distance(
    actual_dataset: Dataset,
    reference_dataset: Dataset,
    column_name: str,
    slicing_function: Optional[SlicingFunction] = None,
    threshold: float = 0.2,
) -> TestResult:
    """Test drift by earth mover’s distance.

    Test if the earth movers distance between the actual and reference datasets
    is below a threshold for a given numerical feature.

    For example, if the threshold is set to 0.1, the test is passed when the
    earth movers distance of the numerical variable between the actual and
    reference datasets is lower than 0.1. It means that we cannot assume drift
    for this variable.

    Parameters
    ----------
    actual_dataset : Dataset
        Actual dataset used to compute the test
    reference_dataset : Dataset
        Reference dataset used to compute the test
    column_name : str
        Name of column with numerical feature
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on both actual and reference datasets
    threshold : float
        Threshold for earth mover's distance. Default is 0.2.

    Returns
    -------
    TestResult
        The test result.
    """
    if slicing_function:
        test_name = "test_drift_earth_movers_distance"
        actual_dataset = actual_dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=actual_dataset, dataset_name="actual_dataset", test_name=test_name)
        reference_dataset = reference_dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=reference_dataset, dataset_name="reference_dataset", test_name=test_name)

    actual_series, reference_series = _extract_series(actual_dataset, reference_dataset, column_name, "numeric")

    metric = _calculate_earth_movers_distance(actual_series, reference_series)

    passed = bool(metric <= threshold)

    messages: Optional[List[TestMessage]] = None

    if not passed:
        messages = [
            TestMessage(
                type=TestMessageLevel.ERROR,
                text=f"The data is drifting (metric is equal to {np.round(metric, 9)} and is below the test risk level {threshold}) ",
            )
        ]
    return TestResult(
        actual_slices_size=[len(actual_series)],
        reference_slices_size=[len(reference_series)],
        passed=True if threshold is None else passed,
        metric=metric,
        messages=messages,
    )


@test(
    name="Label drift (PSI)",
    debug_description=debug_description_prefix + "with <b>the categories that have drifted "
    "the most from the 'actual_dataset'</b>.",
)
def test_drift_prediction_psi(
    model: BaseModel,
    actual_dataset: Dataset,
    reference_dataset: Dataset,
    slicing_function: Optional[SlicingFunction] = None,
    max_categories: int = 10,
    threshold: float = 0.2,
    psi_contribution_percent: float = 0.2,
):
    """Tests drift of predictions by PSI score.

    Test if the PSI score between the reference and actual datasets is below the
    threshold for the classification labels predictions.

    Parameters
    ----------
    model : BaseModel
        Model used to compute the test
    actual_dataset : Dataset
        Actual dataset used to compute the test
    reference_dataset : Dataset
        Reference dataset used to compute the test
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on both actual and reference datasets
    max_categories : int
        The maximum categories to compute the PSI score
    threshold : float
        Threshold value for PSI. Default is 0.2.
    psi_contribution_percent : float
        The ratio between the PSI score of a given category over the total PSI
        score of the categorical variable. If there is a drift, the test
        provides all the categories that have a PSI contribution over than this
        ratio.
    debug : bool
        If True and the test fails, a dataset will be provided containing the
        actual_dataset rows with the categories that have drifted the most (more
        than psi_contribution_percent of the total PSI score).

    Returns
    -------
    TestResult
        The test result.
    """
    if slicing_function:
        test_name = "test_drift_prediction_psi"
        actual_dataset = actual_dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=actual_dataset, dataset_name="actual_dataset", test_name=test_name)
        reference_dataset = reference_dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=reference_dataset, dataset_name="reference_dataset", test_name=test_name)

    prediction_reference = pd.Series(model.predict(reference_dataset).prediction)
    prediction_actual = pd.Series(model.predict(actual_dataset).prediction)
    messages, passed, total_psi, output_data = _test_series_drift_psi(
        prediction_actual,
        prediction_reference,
        "prediction",
        max_categories,
        psi_contribution_percent,
        threshold,
    )

    # --- debug ---
    output_ds = list()
    if not passed:
        check_if_debuggable(actual_dataset, reference_dataset)
        main_drifting_modalities_bool = output_data["Psi"] > psi_contribution_percent * total_psi
        modalities_list = output_data[main_drifting_modalities_bool]["Modality"].tolist()
        if modalities_list:
            filtered_modalities = [w for w in modalities_list if not re.match(other_modalities_pattern, str(w))]
            output_ds.append(
                actual_dataset.slice(
                    lambda df: df.loc[prediction_actual.isin(filtered_modalities).values], row_level=False
                )
            )

            test_name = inspect.stack()[0][3]
            if len(output_ds[0].df.index) == 0:
                raise ValueError(
                    test_name + f": the categories {filtered_modalities} completely drifted as they are not present "
                    f"in the 'actual_dataset'"
                )
    # ---

    return TestResult(
        actual_slices_size=[len(actual_dataset)],
        reference_slices_size=[len(reference_dataset)],
        passed=passed,
        metric=total_psi,
        messages=messages,
        output_ds=output_ds,
    )


def _test_series_drift_psi(
    actual_series,
    reference_series,
    test_data,
    max_categories,
    psi_contribution_percent,
    threshold,
):
    total_psi, output_data = _calculate_drift_psi(actual_series, reference_series, max_categories)
    passed = True if threshold is None else bool(total_psi <= threshold)
    main_drifting_modalities_bool = output_data["Psi"] > psi_contribution_percent * total_psi
    messages = _generate_message_modalities(main_drifting_modalities_bool, output_data, test_data)
    return messages, passed, total_psi, output_data


def _generate_message_modalities(main_drifting_modalities_bool, output_data, test_data):
    modalities_list = output_data[main_drifting_modalities_bool]["Modality"].tolist()
    filtered_modalities = [w for w in modalities_list if not re.match(other_modalities_pattern, str(w))]
    messages: Optional[List[TestMessage]] = None
    if filtered_modalities:
        messages = [
            TestMessage(
                type=TestMessageLevel.ERROR,
                text=f"The {test_data} is drifting for the following modalities: {','.join(str(filtered_modalities))}",
            )
        ]
    return messages


@test(
    name="Label drift (Chi-squared)",
    debug_description=debug_description_prefix + "with <b>the categories that have drifted "
    "the most from the 'actual_dataset'</b>.",
)
def test_drift_prediction_chi_square(
    model: BaseModel,
    actual_dataset: Dataset,
    reference_dataset: Dataset,
    slicing_function: Optional[SlicingFunction] = None,
    max_categories: int = 10,
    threshold: float = 0.05,
    chi_square_contribution_percent: float = 0.2,
):
    """Tests drift of predictions by chi-squared test.

    Test if the chi-square p-value between the reference and actual datasets is
    below the threshold for the classification labels predictions for a given
    slice.

    Parameters
    ----------
    model : BaseModel
        Model used to compute the test
    actual_dataset : Dataset
        Actual dataset used to compute the test
    reference_dataset : Dataset
        Reference dataset used to compute the test
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on both actual and reference datasets
    max_categories : int
        the maximum categories to compute the PSI score.
    threshold : float
        Threshold value of p-value of chi-square. Default is 0.05.
    chi_square_contribution_percent : float
        the ratio between the Chi-Square value of a given category over the
        total chi-square value of the categorical variable. If there is a drift,
        the test provides all the categories that have a PSI contribution over
        than this ratio.
    debug : bool
        If True and the test fails,
        a dataset will be provided containing the actual_dataset rows with the
        categories that have drifted the most (more than
        chi_square_contribution_percent of the total chi squared score).

    Returns
    -------
    TestResult
        The test result.
    """
    if slicing_function:
        test_name = "test_drift_prediction_chi_square"
        actual_dataset = actual_dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=actual_dataset, dataset_name="actual_dataset", test_name=test_name)
        reference_dataset = reference_dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=reference_dataset, dataset_name="reference_dataset", test_name=test_name)

    prediction_reference = pd.Series(model.predict(reference_dataset).prediction)
    prediction_actual = pd.Series(model.predict(actual_dataset).prediction)

    messages, chi_square, p_value, passed, output_data = _test_series_drift_chi(
        prediction_actual,
        prediction_reference,
        "prediction",
        chi_square_contribution_percent,
        max_categories,
        threshold,
    )

    # --- debug ---
    output_ds = list()
    if not passed:
        check_if_debuggable(actual_dataset, reference_dataset)
        main_drifting_modalities_bool = output_data["Chi_square"] > chi_square_contribution_percent * chi_square
        modalities_list = output_data[main_drifting_modalities_bool]["Modality"].tolist()
        if modalities_list:
            filtered_modalities = [w for w in modalities_list if not re.match(other_modalities_pattern, str(w))]
            output_ds.append(
                actual_dataset.slice(
                    lambda df: df.loc[prediction_actual.isin(filtered_modalities).values], row_level=False
                )
            )
    # ---

    return TestResult(
        actual_slices_size=[len(actual_dataset)],
        reference_slices_size=[len(reference_dataset)],
        passed=passed,
        metric=p_value,
        messages=messages,
        output_ds=output_ds,
    )


def _test_series_drift_chi(
    actual_series,
    reference_series,
    test_data,
    chi_square_contribution_percent,
    max_categories,
    threshold,
):
    chi_square, p_value, output_data = _calculate_chi_square(actual_series, reference_series, max_categories)
    passed = bool(p_value > threshold)
    main_drifting_modalities_bool = output_data["Chi_square"] > chi_square_contribution_percent * chi_square
    messages = _generate_message_modalities(main_drifting_modalities_bool, output_data, test_data)
    return messages, chi_square, p_value, passed, output_data


@test(name="Classification Probability drift (Kolmogorov-Smirnov)", tags=["classification"])
@validate_classification_label
def test_drift_prediction_ks(
    model: BaseModel,
    actual_dataset: Dataset,
    reference_dataset: Dataset,
    slicing_function: Optional[SlicingFunction] = None,
    classification_label: Optional[str] = None,
    threshold: Optional[float] = None,
) -> TestResult:
    """Tests drift of predictions by Kolmogorov-Smirnov test.

    Test if the p-value of the KS test for prediction between the reference and
    actual datasets for a given subpopulation is above the threshold.

    Example: The test is passed when the p-value of the KS test for the
    prediction for females between reference and actual dataset is higher than
    0.05. It means that the null hypothesis cannot be rejected at 5% level and
    we cannot assume drift for this variable.

    Parameters
    ----------
    model : BaseModel
        Model used to compute the test
    actual_dataset : Dataset
        Actual dataset used to compute the test
    reference_dataset : Dataset
        Reference dataset used to compute the test
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on both actual and reference datasets
    classification_label : Optional[str]
        One specific label value from the target column for classification model
    threshold : Optional[float]
        Threshold for p-value of Kolmogorov-Smirnov test

    Returns
    -------
    TestResult
        The test result.
    """
    if slicing_function:
        test_name = "test_drift_prediction_ks"
        actual_dataset = actual_dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=actual_dataset, dataset_name="actual_dataset", test_name=test_name)
        reference_dataset = reference_dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=reference_dataset, dataset_name="reference_dataset", test_name=test_name)

    # Try to automatically cast `classification_label` to the right type
    if classification_label is not None and isinstance(model.meta.classification_labels[0], numbers.Number):
        try:
            classification_label = int(classification_label)
        except ValueError:
            pass

    prediction_reference = (
        pd.Series(model.predict(reference_dataset).all_predictions[classification_label].values)
        if model.is_classification
        else pd.Series(model.predict(reference_dataset).prediction)
    )
    prediction_actual = (
        pd.Series(model.predict(actual_dataset).all_predictions[classification_label].values)
        if model.is_classification
        else pd.Series(model.predict(actual_dataset).prediction)
    )

    result: Ks_2sampResult = _calculate_ks(prediction_reference, prediction_actual)

    passed = True if threshold is None else bool(result.pvalue >= threshold)

    messages = _generate_message_ks(passed, result, threshold, "prediction")

    return TestResult(
        actual_slices_size=[len(actual_dataset)],
        reference_slices_size=[len(reference_dataset)],
        passed=passed,
        metric=result.pvalue,
        messages=messages,
    )


def _generate_message_ks(passed, result, threshold, data_type):
    messages: Optional[List[TestMessage]] = None
    if not passed:
        messages = [
            TestMessage(
                type=TestMessageLevel.ERROR,
                text=f"The {data_type} is drifting (p-value is equal to {np.round(result.pvalue, 9)} "
                f"and is below the test risk level {threshold}) ",
            )
        ]
    return messages


@test(name="Classification Probability drift (Earth mover's distance)", tags=["classification"])
@validate_classification_label
def test_drift_prediction_earth_movers_distance(
    model: BaseModel,
    actual_dataset: Dataset,
    reference_dataset: Dataset,
    slicing_function: Optional[SlicingFunction] = None,
    classification_label: Optional[str] = None,
    threshold: float = 0.2,
) -> TestResult:
    """Tests drift of predictions by earth mover’s distance.

    Test if the Earth Mover’s Distance value between the reference and actual
    datasets is below the threshold for the classification labels predictions
    for classification model and prediction for regression models.

    Examples
    For classification: the test is passed when the  Earth Mover’s Distance
    value of classification labels probabilities for females between reference
    and actual sets is below 0.2

    For regression: the test is passed when the  Earth Mover’s Distance value of
    prediction for females between reference and actual sets is below 0.2

    Parameters
    ----------
    model : BaseModel
        Model to test
    actual_dataset : Dataset
        Actual dataset used to compute the test
    reference_dataset : Dataset
        Reference dataset used to compute the test
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on both actual and reference datasets
    classification_label : Optional[str]
        One specific label value from the target column for classification model
    threshold : float
        Threshold for earth mover's distance. Default is 0.2.

    Returns
    -------
    TestResult
        The test result.
    """
    if slicing_function:
        test_name = "test_drift_prediction_earth_movers_distance"
        actual_dataset = actual_dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=actual_dataset, dataset_name="actual_dataset", test_name=test_name)
        reference_dataset = reference_dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=reference_dataset, dataset_name="reference_dataset", test_name=test_name)

    # Try to automatically cast `classification_label` to the right type
    if classification_label is not None and isinstance(model.meta.classification_labels[0], numbers.Number):
        try:
            classification_label = int(classification_label)
        except ValueError:
            pass

    prediction_reference = (
        model.predict(reference_dataset).all_predictions[classification_label].values
        if model.is_classification
        else model.predict(reference_dataset).prediction
    )
    prediction_actual = (
        model.predict(actual_dataset).all_predictions[classification_label].values
        if model.is_classification
        else model.predict(actual_dataset).prediction
    )

    metric = _calculate_earth_movers_distance(prediction_reference, prediction_actual)

    passed = True if threshold is None else bool(metric <= threshold)
    messages: Optional[typing.List[TestMessage]] = None

    if not passed:
        messages = [
            TestMessage(
                type=TestMessageLevel.ERROR,
                text=f"The prediction is drifting (metric is equal to {np.round(metric, 9)} "
                f"and is above the test risk level {threshold}) ",
            )
        ]

    return TestResult(
        actual_slices_size=[len(actual_dataset)],
        reference_slices_size=[len(reference_dataset)],
        passed=bool(True if threshold is None else metric <= threshold),
        metric=metric,
        messages=messages,
    )



from typing import Callable, List, Optional, Set, Union

import copy
import inspect
import pickle
import sys
from abc import ABC, abstractmethod
from pathlib import Path

import yaml

from giskard.core.core import SMT, TestFunctionMeta
from giskard.core.savable import Artifact
from giskard.core.test_result import TestResult
from giskard.core.validation import configured_validate_arguments
from giskard.exceptions.giskard_exception import python_env_exception_helper
from giskard.registry.registry import get_object_uuid, tests_registry
from giskard.registry.utils import dump_by_value
from giskard.utils.analytics_collector import analytics

DATA_PKL = "data.pkl"

Result = Union[TestResult, bool]


class GiskardTest(Artifact[TestFunctionMeta], ABC):
    """
    The base class of all Giskard's tests.

    The tests are executed inside the `execute` method. All arguments should be passed in the `__init__` method. It is
    required to set default values for all arguments (either `None` or a default values).
    """

    def __init__(self):
        test_uuid = get_object_uuid(type(self))
        meta = tests_registry.get_test(test_uuid)
        if meta is None:
            # equivalent to adding @test decorator
            from giskard.registry.decorators import test

            test(type(self))
            meta = tests_registry.get_test(test_uuid)
        super(GiskardTest, self).__init__(meta)

    @abstractmethod
    def execute(self) -> Result:
        """
        Execute the test
        :return: A TestResult or a bool containing detailed information of the test execution results
        :rtype: Union[TestResult, bool]
        """
        ...

    @classmethod
    def _get_name(cls) -> str:
        return "tests"

    @classmethod
    def _get_meta_class(cls) -> type(SMT):
        return TestFunctionMeta

    def _get_uuid(self) -> str:
        return get_object_uuid(type(self))

    def _save_locally(self, local_dir: Path):
        with open(Path(local_dir) / DATA_PKL, "wb") as f:
            dump_by_value(type(self), f)

    @classmethod
    def _load_meta_locally(cls, local_dir, uuid: str) -> Optional[TestFunctionMeta]:
        meta = tests_registry.get_test(uuid)

        if meta is not None:
            return meta

        with open(local_dir / "meta.yaml", "r", encoding="utf-8") as f:
            return yaml.load(f, Loader=yaml.Loader)

    @classmethod
    def load(cls, local_dir: Path, uuid: str, meta: TestFunctionMeta):
        if local_dir.exists():
            with open(Path(local_dir) / "data.pkl", "rb") as f:
                try:
                    func = pickle.load(f)
                except Exception as e:
                    raise python_env_exception_helper(cls.__name__, e)
        elif meta.module in sys.modules and hasattr(sys.modules[meta.module], meta.name):
            func = getattr(sys.modules[meta.module], meta.name)
        else:
            return None

        if inspect.isclass(func) or hasattr(func, "meta"):
            giskard_test = func()
        elif isinstance(func, GiskardTest):
            giskard_test = func
        else:
            giskard_test = GiskardTestMethod(func)

        tests_registry.add_func(meta)
        giskard_test.meta = meta

        return giskard_test

    def assert_(self):
        """Wrap execution of the test into a "assert" for unit test executions."""
        result = self.execute()
        if isinstance(result, bool):
            assert result
        else:
            if result.messages:
                message = " ".join([getattr(message, "text", None) or repr(message) for message in result.messages])
            else:
                # Pass more context in case the message is empty
                message = " ".join(str(result).replace("\n", " ").split())

            assert result.passed, message

    @property
    def display_name(self):
        return self.meta.display_name


Function = Callable[..., Result]

Test = Union[GiskardTest, Function]


class GiskardTestMethod(GiskardTest):
    def __init__(self, test_fn: Function) -> None:
        self.args = list()
        self.kwargs = dict()

        self.is_initialized = False
        self.test_fn = test_fn
        test_uuid = get_object_uuid(self.test_fn)
        meta = tests_registry.get_test(test_uuid)

        if meta is None:
            # equivalent to adding @test decorator
            from giskard.registry.decorators import test

            test()(test_fn)
            meta = tests_registry.get_test(test_uuid)

        super(GiskardTest, self).__init__(meta)

    def __deepcopy__(self, memo):
        # This is necessary, to avoid deepcopy model, which can have thread lock, or some unpickable object
        # https://stackoverflow.com/questions/53601999/getting-typeerror-cant-pickle-thread-lock-objects-when-an-object-is-deepcopi
        # https://stackoverflow.com/questions/1500718/how-to-override-the-copy-deepcopy-operations-for-a-python-object

        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        filtered_items = [(k, v) for k, v in self.__dict__.items() if k not in ["args", "kwargs"]]
        for k, v in filtered_items:
            setattr(result, k, copy.deepcopy(v, memo))
        result.args = list()
        result.kwargs = dict()
        return result

    def __call__(self, *args, **kwargs) -> "GiskardTestMethod":
        instance = copy.deepcopy(self)

        instance.is_initialized = True
        instance.args = args
        instance.kwargs = kwargs

        return instance

    @property
    def dependencies(self) -> Set[Artifact]:
        from inspect import Parameter, signature

        parameters: List[Parameter] = list(signature(self.test_fn).parameters.values())

        return set([param.default for param in parameters if isinstance(param.default, Artifact)])

    @property
    def params(self):
        params = self.kwargs.copy()

        for idx, arg in enumerate(self.args):
            params[next(iter([arg.name for arg in self.meta.args.values() if arg.argOrder == idx]))] = arg

        return params

    def execute(self) -> Result:
        analytics.track("test:execute", {"test_name": self.meta.full_name})

        # if params contains debug then we check if test_fn has debug argument
        if "debug" in self.kwargs and "debug" not in list(inspect.signature(self.test_fn).parameters.keys()):
            self.kwargs.pop("debug")

        return configured_validate_arguments(self.test_fn)(*self.args, **self.kwargs)

    def __repr__(self) -> str:
        if not self.is_initialized:
            hint = "Test object hasn't been initialized, call it by providing the inputs"
        else:
            hint = 'To execute the test call "execute()" method'
        output = [hint]
        if self.params and len(self.params):
            output.append(f"Named inputs: {self.params}")

        return "\n".join(output)

    def _save_locally(self, local_dir: Path):
        with open(Path(local_dir) / "data.pkl", "wb") as f:
            dump_by_value(self.test_fn, f)



from typing import Any, Optional

import itertools
import json
import re
import unicodedata
from pathlib import Path

import numpy as np
import pandas as pd
from num2words import num2words

from ...datasets import Dataset
from ...functions.transformation import gruber
from .base_perturbation_function import PerturbationFunction


class TextTransformation(PerturbationFunction):
    def __init__(self, column: str, needs_dataset: bool = False) -> None:
        super().__init__(column, needs_dataset=needs_dataset)

    def execute(self, data: pd.DataFrame) -> pd.DataFrame:
        feature_data = data[self.column].dropna().astype(str)
        data.loc[feature_data.index, self.column] = feature_data.apply(self.make_perturbation)
        return data


class TextUppercase(TextTransformation):
    name = "Transform to uppercase"

    def execute(self, data: pd.DataFrame) -> pd.DataFrame:
        feature_data = data[self.column].dropna().astype(str)
        data.loc[feature_data.index, self.column] = feature_data.str.upper()
        return data


class TextLowercase(TextTransformation):
    name = "Transform to lowercase"

    def execute(self, data: pd.DataFrame) -> pd.DataFrame:
        feature_data = data[self.column].dropna().astype(str)
        data.loc[feature_data.index, self.column] = feature_data.str.lower()
        return data


class TextTitleCase(TextTransformation):
    name = "Transform to title case"

    def execute(self, data: pd.DataFrame) -> pd.DataFrame:
        feature_data = data[self.column].dropna().astype(str)
        data.loc[feature_data.index, self.column] = feature_data.str.title()
        return data


class TextTypoTransformation(TextTransformation):
    name = "Add typos"

    def __init__(self, column: str, rate: float = 0.05, min_length: int = 10, rng_seed: int = 1729):
        super().__init__(column)
        from .entity_swap import typos

        self.rate = rate
        self.min_length = min_length
        self._key_typos = typos
        self.rng = np.random.default_rng(seed=rng_seed)

    def make_perturbation(self, x: str) -> str:
        # Skip if the text is too short
        if len(x) < self.min_length:
            return x

        # We consider four types of typos:
        # - Insertion
        # - Deletion
        # - Replacement
        # - Transposition

        # Empirical probabilities for each category
        category_prob = [0.2, 0.2, 0.5, 0.1]

        # How many typos we generate and in which positions
        num_typos = self.rng.poisson(self.rate * len(re.sub(r"\s+", "", x)))

        # Are they insertion, deletion, replacement, or transposition?
        pos_cat = self.rng.choice(4, size=num_typos, p=category_prob, replace=True)

        for cat in pos_cat:
            # get a random position
            i = self.rng.integers(0, len(x))

            if cat == 0:  # insertion
                t = self._random_key_typo(x[i])
                x = x[:i] + t + x[i:]
            elif cat == 1:  # deletion
                if x[i].isspace():  # don’t delete spaces
                    i = min(i + 1, len(x) - 1)
                x = x[:i] + x[i + 1 :]
            elif cat == 2:  # replacement
                x = x[:i] + self._random_key_typo(x[i]) + x[i + 1 :]
            else:  # transposition
                if i < len(x) - 1:
                    x = x[:i] + x[i + 1] + x[i] + x[i + 2 :]
        return x

    def _random_key_typo(self, char: str):
        if char.lower() in self._key_typos:
            typo = self.rng.choice(self._key_typos[char.lower()])
            return typo if char.islower() else typo.upper()
        return char


class TextFromOCRTypoTransformation(TextTransformation):
    name = "Add typos from OCR"

    def __init__(self, column: str, rate: float = 0.05, min_length: int = 10, rng_seed: int = 1729):
        super().__init__(column)
        from .entity_swap import ocr_typos

        self.rate = rate
        self.min_length = min_length
        self._ocr_typos = ocr_typos
        self.rng = np.random.default_rng(seed=rng_seed)

    def make_perturbation(self, x: str) -> str:
        # Check if the input is None
        if x is None:
            return None
        # Skip if the text is too short
        if len(x) < self.min_length:
            return x

        # OCR typos consist only of replacements and deletions
        category_prob = [0.8, 0.2]  # Probability [replacement, deletion]

        # How many typos to introduce
        num_typos = self.rng.poisson(self.rate * len(re.sub(r"\s+", "", x)))

        # Are they replacement or deletion?
        pos_cat = self.rng.choice(2, size=num_typos, p=category_prob, replace=True)

        for cat in pos_cat:
            # Get a random position avoiding spaces for deletion
            i = self.rng.integers(0, len(x))
            if cat == 0:  # Replacement
                if x[i].lower() in self._ocr_typos:
                    x = x[:i] + self._random_ocr_typo(x[i]) + x[i + 1 :]
            elif cat == 1:  # Deletion
                if not x[i].isspace():  # Don’t delete spaces
                    x = x[:i] + x[i + 1 :]
        return x

    def _random_ocr_typo(self, char: str) -> str:
        if char.lower() in self._ocr_typos:
            typo = self.rng.choice(self._ocr_typos[char.lower()])
            return typo if char.islower() else typo.upper()
        return char


class TextPunctuationRemovalTransformation(TextTransformation):
    name = "Punctuation Removal"

    _punctuation = "¡!¿?⸘‽“”‘’‛‟.,‚„'\"′″´˝`…:;­–‑—§¶†‡/-‒‼︎⁇⁈⁉︎❛❜❝❞"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._trans_table = str.maketrans("", "", self._punctuation)
        self._regex = re.compile(rf"\b[{re.escape(self._punctuation)}]+\b")

    def make_perturbation(self, text: str) -> str:
        # Split URLs so that they are not affected by the transformation
        pieces = gruber.split(text)

        # Non-URLs are even-numbered entries
        for i in range(0, len(pieces), 2):
            # Replace punctuation at boundaries with a space, so that we don't
            # accidentally join words together. Then strip the rest.
            pieces[i] = self._regex.sub(" ", pieces[i]).translate(self._trans_table)

        return "".join(pieces)


class TextAccentRemovalTransformation(TextTransformation):
    name = "Accent Removal"

    def __init__(self, column: str, rate: float = 1.0, rng_seed: int = 1729):
        super().__init__(column)
        self.rate = rate
        self.rng = np.random.default_rng(seed=rng_seed)

    def make_perturbation(self, text: str) -> str:
        return "".join(
            char
            for char in unicodedata.normalize("NFD", text)
            if unicodedata.category(char) != "Mn" or self.rng.random() > self.rate
        )


class TextLanguageBasedTransformation(TextTransformation):
    def __init__(self, column: str, rng_seed: int = 1729):
        super().__init__(column, needs_dataset=True)
        self._lang_dictionary = dict()
        self._load_dictionaries()
        self.rng = np.random.default_rng(seed=rng_seed)

    def _load_dictionaries(self):
        raise NotImplementedError()

    def execute(self, dataset: Dataset) -> pd.DataFrame:
        feature_data = dataset.df.loc[:, (self.column,)].dropna().astype(str)
        meta_dataframe = dataset.column_meta[self.column, "text"].add_suffix("__gsk__meta")
        feature_data = feature_data.join(meta_dataframe)
        dataset.df.loc[feature_data.index, self.column] = feature_data.apply(self.make_perturbation, axis=1)
        return dataset.df

    def make_perturbation(self, row: pd.Series) -> Any:
        raise NotImplementedError()

    def _switch(self, word: str, language: str) -> Optional[tuple[str, str]]:
        raise NotImplementedError()

    def _select_dict(self, language: str) -> Optional[Any]:
        try:
            return self._lang_dictionary[language]
        except KeyError:
            return None


class TextGenderTransformation(TextLanguageBasedTransformation):
    name = "Switch Gender"

    def _load_dictionaries(self):
        from .entity_swap import gender_switch_en, gender_switch_fr

        self._lang_dictionary = {"en": gender_switch_en, "fr": gender_switch_fr}

    def make_perturbation(self, row: pd.Series) -> str:
        text = row[self.column]
        language = row["language__gsk__meta"]

        if language not in self._lang_dictionary:
            return text

        replacements = [self._switch(token, language) for token in text.split()]
        replacements = [r for r in replacements if r is not None]

        new_text = text
        for original_word, switched_word in replacements:
            new_text = re.sub(rf"\b{re.escape(original_word)}\b", switched_word, new_text)

        return new_text

    def _switch(self, word: str, language: str) -> Optional[tuple[str, str]]:
        try:
            return (word, self._lang_dictionary[language][word.lower()])
        except KeyError:
            return None


class TextNumberToWordTransformation(TextLanguageBasedTransformation):
    name = "Transform numbers to words"

    def _load_dictionaries(self):
        # Regex to match numbers in text
        self._regex = re.compile(r"(?<!\d/)(?<!\d\.)\b\d+(?:\.\d+)?\b(?!(?:\.\d+)?@|\d?/?\d)")

    def make_perturbation(self, row: pd.Series) -> str:
        # Replace numbers with words
        value = row[self.column]
        if pd.isna(value):
            return value
        lang = row["language__gsk__meta"] if not pd.isna(row["language__gsk__meta"]) else "en"

        if lang == "fa" or lang == "id":
            # In num2words, the convertor of "fa" and "id" are buggy,
            # see https://github.com/savoirfairelinux/num2words/issues/476
            # Give up doing this now, wait for merging https://github.com/savoirfairelinux/num2words/pull/524
            return value

        try:
            return self._regex.sub(lambda x: num2words(x.group(), lang=lang), value)
        except NotImplementedError:
            # Fallback to english in case of unimplemented
            return self._regex.sub(lambda x: num2words(x.group(), lang="en"), value)


class TextReligionTransformation(TextLanguageBasedTransformation):
    name = "Switch Religion"

    def _load_dictionaries(self):
        from .entity_swap import religion_dict_en, religion_dict_fr

        self._lang_dictionary = {"en": religion_dict_en, "fr": religion_dict_fr}

    def make_perturbation(self, row: pd.Series) -> str:
        # Get text
        text = row[self.column]

        # Get language and corresponding dictionary
        language = row["language__gsk__meta"]
        religion_dict = self._select_dict(language)

        # Check if we support this language
        if religion_dict is None:
            return text

        # Mask entities and prepare replacements
        replacements = []
        for n_list, term_list in enumerate(religion_dict):
            for n_term, term in enumerate(term_list):
                mask_value = f"__GSK__ENT__RELIGION__{n_list}__{n_term}__"
                text, num_rep = re.subn(rf"\b{re.escape(term)}(s?)\b", rf"{mask_value}\1", text, flags=re.IGNORECASE)
                if num_rep > 0:
                    i = (n_term + 1 + self.rng.choice(len(term_list) - 1)) % len(term_list)
                    replacement = term_list[i]
                    replacements.append((mask_value, replacement))

        # Replace masks
        for mask, replacement in replacements:
            text = text.replace(mask, replacement)

        return text


class TextNationalityTransformation(TextLanguageBasedTransformation):
    name = "Switch countries from high- to low-income and vice versa"

    def _load_dictionaries(self):
        with Path(__file__).parent.joinpath("nationalities.json").open("r", encoding="utf-8") as f:
            nationalities_dict = json.load(f)
        self._lang_dictionary = {"en": nationalities_dict["en"], "fr": nationalities_dict["fr"]}

    def make_perturbation(self, row: pd.Series) -> str:
        text = row[self.column]
        language = row["language__gsk__meta"]
        nationalities_word_dict = self._select_dict(language)

        if nationalities_word_dict is None:
            return text

        ent_types = list(itertools.product(["country", "nationality"], ["high-income", "low-income"]))

        # Mask entities and prepare replacements
        replacements = []
        for entity_type, income_type in ent_types:
            for n, entity_word in enumerate(nationalities_word_dict[entity_type][income_type]):
                mask_value = f"__GSK__ENT__{entity_type}__{income_type}__{n}__"
                text, num_rep = re.subn(
                    rf"(\W|^){re.escape(entity_word)}(\W|$)",
                    rf"\g<1>{mask_value}\g<2>",
                    text,
                    flags=re.IGNORECASE,
                )
                if num_rep > 0:
                    r_income_type = "low-income" if income_type == "high-income" else "high-income"
                    replacement = self.rng.choice(nationalities_word_dict[entity_type][r_income_type])
                    replacements.append((mask_value, replacement))

        # Replace masks
        for mask, replacement in replacements:
            text = text.replace(mask, replacement)

        return text


class TextFromSpeechTypoTransformation(TextLanguageBasedTransformation):
    name = "Add text from speech typos"

    def __init__(self, column: str, rng_seed: int = 1729, min_length: int = 10):
        super().__init__(column, rng_seed=rng_seed)

        self.min_length = min_length

    def _load_dictionaries(self):
        from .entity_swap import speech_typos

        self._word_typos = speech_typos

    def make_perturbation(self, row: pd.Series) -> str:
        text = row[self.column]
        language = row["language__gsk__meta"]

        # Skip if the text is too short
        if len(text) < self.min_length:
            return text
        # Skip if language isn't supported
        if language != "en":
            return text

        # We are considering homophones
        # Split the input text by spaces to get the words
        words = text.split()
        transformed_words = []

        # Iterate over each word in the input text
        for word in words:
            # Normalize the word to handle cases
            normalized_word = word.lower()

            # Check if the current word is present in _word_typos dictionary
            if normalized_word in self._word_typos:
                # Choose a random typo for the current word
                typo_options = self._word_typos[normalized_word]
                chosen_typo = self.rng.choice(typo_options)

                # Retain original word case
                if word.istitle():
                    chosen_typo = chosen_typo.capitalize()
                elif word.isupper():
                    chosen_typo = chosen_typo.upper()
                else:
                    # Keep typo as it is in the typo dictionary
                    pass

                # Append the typo to the transformed words list
                transformed_words.append(chosen_typo)
            else:
                # The word is not in the dictionary, keep it as is
                transformed_words.append(word)

        # Reconstruct the transformed text from the words
        transformed_text = " ".join(transformed_words)
        return transformed_text



import logging

from ..knowledge_base import KnowledgeBase
from ..testset import QuestionSample
from .base import _BaseModifierGenerator
from .prompt import QAGenerationPrompt
from .simple_questions import SimpleQuestionsGenerator

logger = logging.getLogger(__name__)


SITUATIONAL_CONTEXT_SYSTEM_PROMPT = """You are a creative writer and you goal is to write short situational description of a user using the chatbot and its intentions.
Your description must be 1 sentence long. 
You will be given the chatbot description delimited by XML tags. 
You will also be provided with a context paragraph delimited with <context></context> tags.

The situational description must be plausible given the context and the chatbot description. 
Do not mention that the user is using a chatbot or any thing related to the chatbot description.
If not enough information is provided inside the context use only the chatbot description to generate the situational context.
Be creative and imaginative.

You must only return one the situational description, nothing else.
"""

SITUATIONAL_CONTEXT_INPUT = """
<description>{agent_description}</description>
<context>{context}</context>
"""

SITUATIONAL_QUESTION_SYSTEM_PROMPT = """You are an expert at rewriting questions.
Your task is to re-write questions that will be used to evaluate the following agent:
- Agent description: {agent_description}  

Your task is to add situational context about the user inside the question. 
Please respect the following rules to generate the question:
- The question must include the information from the situational context.
- The question must sound plausible and coming from a real human user.
- The question can start with any form of greetings or not, choose randomly
- The original question and answer should be preserved.
- The question must be self-contained and understandable by humans. 
- The question must be in this language: {language}.

You will be provided the question and answer delimited with XML tags.
You will also be provided a situational context delimited with <situation></situation> tags.
You will return the reformulated question as a single JSON object, with the key 'question'. Make sure you return a valid JSON object.
"""

SITUATIONAL_QUESTION_USER_TEMPLATE = """<question>
{question}
</question>
<answer>
{answer}
</answer>
<situation>
{situation}
</situation>"""

SITUATIONAL_QUESTION_EXAMPLE_INPUT = SITUATIONAL_QUESTION_USER_TEMPLATE.format(
    question="How many people are in France?",
    answer="There are nearly 70M people living in France",
    situation="A student from Canada is going to study in Europe.",
)

SITUATIONAL_QUESTION_EXAMPLE_OUTPUT = (
    """{"question": "I am a student from Canada and I am going to study in Europe. How many people are in France?"}"""
)


class SituationalQuestionsGenerator(_BaseModifierGenerator):
    _base_generator = SimpleQuestionsGenerator()

    _situation_generation_prompt = QAGenerationPrompt(
        system_prompt=SITUATIONAL_CONTEXT_SYSTEM_PROMPT,
        user_input_template=SITUATIONAL_CONTEXT_INPUT,
        example_input=SITUATIONAL_QUESTION_EXAMPLE_INPUT,
        example_output=SITUATIONAL_QUESTION_EXAMPLE_OUTPUT,
    )

    _prompt = QAGenerationPrompt(
        system_prompt=SITUATIONAL_QUESTION_SYSTEM_PROMPT,
        user_input_template=SITUATIONAL_QUESTION_USER_TEMPLATE,
    )

    _question_type = "situational"

    def _modify_question(
        self, question: QuestionSample, knowledge_base: KnowledgeBase, agent_description: str, language: str
    ) -> QuestionSample:
        situation_generation_messages = self._situation_generation_prompt.to_messages(
            system_prompt_input={},
            user_input={
                "agent_description": agent_description,
                "context": question.reference_context,
            },
        )

        situational_context = "A user of the chatbot asks a general question."
        try:
            situational_context = self._llm_client.complete(messages=situation_generation_messages).content
        except Exception as e:
            logger.warning(
                f"Encountered error in situational context generation: {e}. Using default situational context instead."
            )

        messages = self._prompt.to_messages(
            system_prompt_input={
                "agent_description": agent_description,
                "language": language,
            },
            user_input={
                "question": question.question,
                "answer": question.reference_answer,
                "situation": situational_context,
            },
        )

        out = self._llm_complete(messages=messages)
        question.question = out["question"]

        question.metadata["question_type"] = self._question_type
        question.metadata["situational_context"] = situational_context

        return question


situational_questions = SituationalQuestionsGenerator()



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import numpy as np

from giskard.llm.talk.config import ToolDescription
from giskard.llm.talk.tools.predict import PredictTool
from giskard.llm.talk.utils.shap import explain_with_shap

if TYPE_CHECKING:
    from giskard.datasets.base import Dataset
    from giskard.models.base import BaseModel


class SHAPExplanationTool(PredictTool):
    """SHAP explanation Tool.

    Attributes
    ----------
    default_name : str
        The default name of the Tool. Can be re-defined with constructor.
    default_description: str
        The default description of the Tool's functioning. Can be re-defined with constructor.
    """

    default_name: str = "shap_explanation"
    default_description: str = ToolDescription.SHAP_EXPLANATION.value
    _output_template: str = "'{feature_name}' | {attributions_values}"

    def __init__(
        self, model: BaseModel, dataset: Dataset, name: Optional[str] = None, description: Optional[str] = None
    ):
        """Constructor of the class.

        Parameters
        ----------
        model : BaseModel
            The Giskard Model.
        dataset : Dataset
            The Giskard Dataset.
        name : str, optional
            The name of the Tool.
            If not set, the `default_name` is used.
        description : str, optional
            The description of the Tool.
            If not set, the `default_description` is used.
        """
        super().__init__(model, dataset, name, description)  # To reflect the docstring.

    @staticmethod
    def _finalise_output(prediction_result: str, shap_result: np.ndarray) -> str:
        """Define the output format.

        Create Tool's output string, using predictions and related SHAP explanations.

        Parameters
        ----------
        prediction_result : str
            The prediction result.
        shap_result : np.ndarray
            The SHAP explanations.

        Returns
        -------
        str
            The formatted Tool's output.
        """
        shap_result = "\n".join(shap_result)
        return (
            f"Prediction result:\n"
            f"{prediction_result}\n\n"
            f"SHAP result: \n"
            f"'Feature name' | [('SHAP value', 'Feature value'), ...]\n"
            f"-------------------------------------------------------\n"
            f"{shap_result}\n\n"
        )

    def __call__(self, features_dict: dict[str, any]) -> str:
        """Execute the Tool's functionality.

        Calculate the SHAP explanations.

        Parameters
        ----------
        features_dict : dict[str, any]
            The dictionary with features and related values extracted from the query.

        Returns
        -------
        str
            SHAP explanation of the predictions.
        """
        # Get a prediction result from the parent tool, to add more context to the result.
        prediction_result = super().__call__(features_dict)

        model_input = self._prepare_input(features_dict)
        explanations = explain_with_shap(model_input, self._model, self._dataset)

        # Finalise the result.
        shap_result = [
            self._output_template.format(
                feature_name=f_name,
                attributions_values=list(zip(np.abs(explanations[:, f_name].values), explanations[:, f_name].data)),
            )
            for f_name in explanations.feature_names
        ]
        return self._finalise_output(prediction_result, shap_result)



import numpy as np
import pandas as pd
from sklearn.tree import DecisionTreeRegressor

from .slice import GreaterThan, LowerThan, Query, QueryBasedSliceFunction
from .tree_slicer import DecisionTreeSlicer


class MultiscaleSlicer(DecisionTreeSlicer):
    def find_slices(self, features, target=None):
        target = target or self.target
        data = self.dataset.df.loc[:, list(features) + [target]].dropna()
        min_leaf_size = 30

        if len(data) < 2 * min_leaf_size:
            return []

        # Adaptive binning via decision tree
        criterion = self._choose_tree_criterion(data.loc[:, target].values)
        dt = DecisionTreeRegressor(criterion=criterion, min_samples_leaf=min_leaf_size)
        dt.fit(data.loc[:, features], data[target])
        tree = dt.tree_

        # Global median
        global_med = data[target].median()

        # Start from the leaves and go up level by level
        dp = dt.decision_path(data.loc[:, features])

        leaf_bins = tree.apply(data.loc[:, features].values.astype(np.float32))
        dx_agg = data.loc[:, target].groupby(leaf_bins).median()
        deviant_leaves = dx_agg.index[dx_agg > global_med * (1 + self.min_deviation)]

        def get_parent(node):
            lp = np.nonzero(tree.children_left == node)[0]
            rp = np.nonzero(tree.children_right == node)[0]

            return np.concatenate((lp, rp))[0]

        def is_significant_slice(samples):
            return samples[target].median() > (1 + self.min_deviation) * global_med

        selection_nodes = []
        for node_id in deviant_leaves:
            depth = 0

            while node_id is not None:
                parent_id = get_parent(node_id)
                samples_idx = np.nonzero(dp[:, parent_id])[0]

                if is_significant_slice(data.iloc[samples_idx]):
                    node_id = parent_id
                    depth += 1
                else:
                    selection_nodes.append(
                        {
                            "node": node_id,
                            "depth": depth,
                        }
                    )
                    break

        if len(selection_nodes) == 0:
            return []

        # Keep slices that are consistent across 5 levels
        dx_sel = pd.DataFrame(selection_nodes).query("depth >= 3")

        # Reconstruct rules
        slices = []
        for node_id in dx_sel.node.unique():
            th_min = -float("inf")
            th_max = float("inf")

            while node_id != 0:
                parent_id = get_parent(node_id)

                th = tree.threshold[parent_id]
                if tree.children_left[parent_id] == node_id:
                    th_max = min(th_max, th)
                else:
                    th_min = max(th_min, th)

                node_id = parent_id

            clauses = []
            if np.isfinite(th_min):
                clauses.append(GreaterThan(features[0], th_min))
            if np.isfinite(th_max):
                clauses.append(LowerThan(features[0], th_max, True))

            slices.append(QueryBasedSliceFunction(Query(clauses)))

        return slices



import warnings

import numpy as np
import pandas as pd
import pytest

from giskard.core.dataset_validation import validate_optional_target
from giskard.datasets.base import Dataset

MANDATORY_FIELDS = [
    "id",
    "originalSizeBytes",
    "numberOfRows",
    "columnTypes",
    "columnDtypes",
    "compressedSizeBytes",
    "categoryFeatures",
    "createdDate",
]
OPTIONAL_FIELDS = [
    "name",
    "target",
]

valid_df = pd.DataFrame(
    {
        "categorical_column": ["turtle", "crocodile", "turtle"],
        "text_column": ["named Giskard", "a nile crocodile", "etc"],
        "numeric_column": [15.5, 25.9, 2.4],
    }
)
valid_df_column_types = {
    "categorical_column": "category",
    "text_column": "text",
    "numeric_column": "numeric",
}

nonvalid_df = pd.DataFrame(
    {
        "categorical_column": [["turtle"], ["crocodile"], ["turtle"]],
        "text_column": [{1: "named Giskard"}, {2: "a nile crocodile"}, {3: "etc"}],
        "numeric_column": [(15.5, 1), (25.9, 2), (2.4, 3)],
    }
)


def test_factory():
    my_dataset = Dataset(valid_df)
    assert isinstance(my_dataset, Dataset)


def test_validate_optional_target():
    with pytest.warns(
        UserWarning,
        match=r"You did not provide the optional argument 'target'\. 'target' is the column name "
        r"in df corresponding to the actual target variable \(ground truth\)\.",
    ):
        my_dataset = Dataset(valid_df)
        validate_optional_target(my_dataset)

    # https://docs.pytest.org/en/latest/how-to/capture-warnings.html#additional-use-cases-of-warnings-in-tests
    with warnings.catch_warnings():
        warnings.simplefilter("error")

        my_dataset = Dataset(valid_df, target=None)
        validate_optional_target(my_dataset)

        my_dataset = Dataset(valid_df, target="text_column")
        validate_optional_target(my_dataset)


def test_valid_df_column_types():
    # Option 0: none of column_types, cat_columns, infer_column_types = True are provided
    my_dataset = Dataset(valid_df)
    assert my_dataset.column_types == {
        "categorical_column": "category",
        "text_column": "text",
        "numeric_column": "numeric",
    }

    # Option 1: column_types is provided
    my_dataset = Dataset(valid_df, column_types=valid_df_column_types)
    assert my_dataset.column_types == valid_df_column_types

    # Option 2: cat_columns is provided
    cat_columns = ["categorical_column"]
    my_dataset = Dataset(valid_df, cat_columns=cat_columns)
    assert my_dataset.column_types == valid_df_column_types

    # Option 3: infer_column_types is provided
    my_dataset = Dataset(valid_df)
    assert my_dataset.column_types == valid_df_column_types


def test_nonvalid_df_column_types():
    # Option 0: none of column_types, cat_columns, infer_column_types = True are provided
    with pytest.raises(
        TypeError,
        match=r"The following columns in your df: \['categorical_column', 'text_column'\] are not hashable\. "
        r"We currently support only hashable column types such as int, bool, str, tuple and not list or dict\.",
    ):
        Dataset(nonvalid_df)

    # Option 1: column_types is provided
    with pytest.raises(
        TypeError,
        match=r"The following columns in your df: \['categorical_column', 'text_column'\] are not hashable\. "
        r"We currently support only hashable column types such as int, bool, str, tuple and not list or dict\.",
    ):
        Dataset(nonvalid_df, column_types=valid_df_column_types)

    # Option 2: cat_columns is provided
    cat_columns = ["categorical_column"]
    with pytest.raises(
        TypeError,
        match=r"The following columns in your df: \['categorical_column', 'text_column'\] are not hashable\. "
        r"We currently support only hashable column types such as int, bool, str, tuple and not list or dict\.",
    ):
        Dataset(nonvalid_df, cat_columns=cat_columns)

    # Option 3: infer_column_types is provided
    with pytest.raises(
        TypeError,
        match=r"The following columns in your df: \['categorical_column', 'text_column'\] are not hashable\. "
        r"We currently support only hashable column types such as int, bool, str, tuple and not list or dict\.",
    ):
        Dataset(nonvalid_df)


def test_dataset_raises_exception_if_mixed_column_types():
    df = pd.DataFrame({"feature": [1, 2, "string", None, np.nan], "target": [0, 0, 1, 1, 0]})

    with pytest.raises(TypeError):
        Dataset(df, target="target")


def test_inference_priority():
    column_types = {"categorical_column": "text"}
    expected_df_column_types = {
        "categorical_column": "text",
        "text_column": "text",
        "numeric_column": "numeric",
    }

    # Case 1: only one column in column_types is provided
    my_dataset = Dataset(valid_df, column_types=column_types)
    assert my_dataset.column_types == expected_df_column_types

    # Case 2: one column in column_types is provided, one in cat_columns
    my_dataset = Dataset(valid_df, column_types=column_types, cat_columns=["categorical_column"])
    assert my_dataset.column_types == valid_df_column_types

    # Case 3: an unknown column in column_types is provided
    column_types = {"unknown_column": "text"}
    my_dataset = Dataset(valid_df, column_types=column_types, cat_columns=["categorical_column"])
    assert my_dataset.column_types == valid_df_column_types


def test_numeric_column_names():
    df = pd.DataFrame(np.ones((10, 3)), columns=[1, 2, 3])

    assert Dataset(df, target=2)
    assert Dataset(df, column_types={1: "numeric"})


def test_infer_column_types():
    # if df_size >= 100     ==> category_threshold = floor(log10(df_size))
    assert Dataset(pd.DataFrame({"f": [1, 2] * 50})).column_types["f"] == "category"
    assert Dataset(pd.DataFrame({"f": ["a", "b"] * 50})).column_types["f"] == "category"
    assert Dataset(pd.DataFrame({"f": ["a", "b"] * 49})).column_types["f"] == "category"
    assert Dataset(pd.DataFrame({"f": [1, 2, 3] * 50})).column_types["f"] == "numeric"
    assert Dataset(pd.DataFrame({"f": ["a", "b", "c"] * 50})).column_types["f"] == "text"

    # if 2 < df_size < 100  ==> category_threshold = 2
    assert Dataset(pd.DataFrame({"f": [1, 2, 1]})).column_types["f"] == "category"
    assert Dataset(pd.DataFrame({"f": ["a", "b", "a"]})).column_types["f"] == "category"
    assert Dataset(pd.DataFrame({"f": [1, 2, 3]})).column_types["f"] == "numeric"
    assert Dataset(pd.DataFrame({"f": ["a", "b", "c"]})).column_types["f"] == "text"

    # if df_size <= 2       ==> category_threshold = 0 (column is text)
    assert Dataset(pd.DataFrame({"f": [1, 2]})).column_types["f"] == "numeric"
    assert Dataset(pd.DataFrame({"f": ["a", "b"]})).column_types["f"] == "text"
    assert Dataset(pd.DataFrame({"f": [1]})).column_types["f"] == "numeric"
    assert Dataset(pd.DataFrame({"f": ["a"]})).column_types["f"] == "text"



from typing import Optional, Sequence

from abc import ABC, abstractmethod

import numpy as np
import pandas as pd

from ...datasets.base import Dataset
from ...llm import LLMImportError
from ...models.base import BaseModel
from ...models.base.model_prediction import ModelPredictionResults
from ..issues import Issue, IssueLevel, Robustness
from ..logger import logger
from ..registry import Detector
from .base_perturbation_function import PerturbationFunction
from .numerical_transformations import NumericalTransformation
from .text_transformations import TextTransformation


def _relative_delta(actual: np.ndarray, reference: np.ndarray) -> np.ndarray:
    """
    Computes elementwise relative delta. If reference[i] == 0, we replace it with epsilon
    to avoid division by zero.
    """
    epsilon = 1e-9
    safe_ref = np.where(reference == 0, epsilon, reference)
    return (actual - reference) / safe_ref


def _get_default_num_samples(model) -> int:
    if model.is_text_generation:
        return 10
    return 1_000


def _get_default_output_sensitivity(model) -> float:
    if model.is_text_generation:
        return 0.15
    return 0.05


def _get_default_threshold(model) -> float:
    if model.is_text_generation:
        return 0.10
    return 0.05


def _generate_robustness_tests(issue: Issue):
    from ...testing.tests.metamorphic import test_metamorphic_invariance

    # Only generates a single metamorphic test
    return {
        f"Invariance to “{issue.transformation_fn}”": test_metamorphic_invariance(
            transformation_function=issue.transformation_fn,
            slicing_function=None,
            threshold=1 - issue.meta["threshold"],
            output_sensitivity=issue.meta.get("output_sentitivity", None),
        )
    }


class BasePerturbationDetector(Detector, ABC):
    """
    Common parent class for metamorphic perturbation detectors (both text and numerical).
    """

    _issue_group = Robustness
    _taxonomy = ["avid-effect:performance:P0201"]

    def __init__(
        self,
        transformations: Optional[Sequence[PerturbationFunction]] = None,
        threshold: Optional[float] = None,
        output_sensitivity: Optional[float] = None,
        num_samples: Optional[int] = None,
    ):
        """
        Creates a new instance of the detector.

        Parameters
        ----------
        transformations: Optional[Sequence[PerturbationFunction]]
            The transformations used in the metamorphic testing. See :ref:`transformation_functions` for details
            about the available transformations. If not provided, a default set of transformations will be used.
        threshold: Optional[float]
            The threshold for the fail rate, which is defined as the proportion of samples for which the model
            prediction has changed. If the fail rate is greater than the threshold, an issue is created.
            If not provided, a default threshold will be used.
        output_sensitivity: Optional[float]
            For regression models, the output sensitivity is the maximum relative change in the prediction that is
            considered acceptable. If the relative change is greater than the output sensitivity, an issue is created.
            This parameter is ignored for classification models. If not provided, a default output sensitivity will be
            used.
        num_samples: Optional[int]
            The maximum number of samples to use for the metamorphic testing. If not provided, a default number of
            samples will be used.
        """
        self.transformations = transformations
        self.threshold = threshold
        self.num_samples = num_samples
        self.output_sensitivity = output_sensitivity

    @abstractmethod
    def _select_features(self, dataset: Dataset, features: Sequence[str]) -> Sequence[str]:
        raise NotImplementedError

    @abstractmethod
    def _get_default_transformations(self) -> Sequence[PerturbationFunction]:
        raise NotImplementedError

    @abstractmethod
    def _supports_text_generation(self) -> bool:
        raise NotImplementedError

    def _compute_passed(
        self,
        model: BaseModel,
        original_pred: ModelPredictionResults,
        perturbed_pred: ModelPredictionResults,
        output_sensitivity: float,
    ) -> np.ndarray:
        if model.is_classification:
            return original_pred.raw_prediction == perturbed_pred.raw_prediction

        elif model.is_regression:
            rel_delta = _relative_delta(perturbed_pred.raw_prediction, original_pred.raw_prediction)
            return np.abs(rel_delta) < output_sensitivity

        elif model.is_text_generation:
            if not self._supports_text_generation():
                raise NotImplementedError("Text generation is not supported by this detector.")
            try:
                import evaluate
            except ImportError as err:
                raise LLMImportError() from err

            scorer = evaluate.load("bertscore")
            score = scorer.compute(
                predictions=perturbed_pred.prediction,
                references=original_pred.prediction,
                model_type="distilbert-base-multilingual-cased",
                idf=True,
            )
            return np.array(score["f1"]) > 1 - output_sensitivity

        else:
            raise NotImplementedError("Only classification, regression, or text generation models are supported.")

    def _create_examples(
        self,
        original_data: Dataset,
        original_pred: ModelPredictionResults,
        perturbed_data: Dataset,
        perturbed_pred: ModelPredictionResults,
        feature: str,
        passed: np.ndarray,
        model: BaseModel,
        transformation_fn,
    ) -> pd.DataFrame:
        examples = original_data.df.loc[~passed, [feature]].copy()
        examples[f"{transformation_fn.name}({feature})"] = perturbed_data.df.loc[~passed, feature]

        examples["Original prediction"] = original_pred.prediction[~passed]
        examples["Prediction after perturbation"] = perturbed_pred.prediction[~passed]

        if model.is_classification:
            examples["Original prediction"] = examples["Original prediction"].astype(str)
            examples["Prediction after perturbation"] = examples["Prediction after perturbation"].astype(str)

            ps_before = pd.Series(original_pred.probabilities[~passed], index=examples.index)
            ps_after = pd.Series(perturbed_pred.probabilities[~passed], index=examples.index)

            examples["Original prediction"] += ps_before.apply(lambda p: f" (p={p:.2f})")
            examples["Prediction after perturbation"] += ps_after.apply(lambda p: f" (p={p:.2f})")

        return examples

    def _detect_issues(
        self,
        model: BaseModel,
        dataset: Dataset,
        transformation,
        features: Sequence[str],
    ) -> Sequence[Issue]:
        # Fall back to defaults if not explicitly set
        num_samples = self.num_samples if self.num_samples is not None else _get_default_num_samples(model)
        threshold = self.threshold if self.threshold is not None else _get_default_threshold(model)
        output_sensitivity = (
            self.output_sensitivity if self.output_sensitivity is not None else _get_default_output_sensitivity(model)
        )

        issues = []
        for feature in features:
            # Build transformation function for this feature
            transformation_fn = transformation(column=feature)
            transformed = dataset.transform(transformation_fn)

            # Select only the records which were changed
            changed_idx = dataset.df.index[transformed.df[feature] != dataset.df[feature]]
            if changed_idx.empty:
                continue

            # Select a random subset of the changed records
            if len(changed_idx) > num_samples:
                rng = np.random.default_rng(747)
                changed_idx = changed_idx[rng.choice(len(changed_idx), num_samples, replace=False)]

            # Build original vs. perturbed datasets
            original_data = Dataset(
                dataset.df.loc[changed_idx],
                target=dataset.target,
                column_types=dataset.column_types,
                validation=False,
            )
            perturbed_data = Dataset(
                transformed.df.loc[changed_idx],
                target=dataset.target,
                column_types=dataset.column_types,
                validation=False,
            )

            # Calculate predictions
            original_pred = model.predict(original_data)
            perturbed_pred = model.predict(perturbed_data)

            passed = self._compute_passed(
                model=model,
                original_pred=original_pred,
                perturbed_pred=perturbed_pred,
                output_sensitivity=output_sensitivity,
            )

            pass_rate = passed.mean()
            fail_rate = 1 - pass_rate
            logger.info(
                f"{self.__class__.__name__}: Testing `{feature}` for perturbation `{transformation.name}`\tFail rate: {fail_rate:.3f}"
            )

            if fail_rate >= threshold:
                # Severity
                issue_level = IssueLevel.MAJOR if fail_rate >= 2 * threshold else IssueLevel.MEDIUM

                # Description
                desc = (
                    "When feature “{feature}” is perturbed with the transformation “{transformation_fn}”, "
                    "the model changes its prediction in {fail_rate_percent}% of the cases. "
                    "We expected the predictions not to be affected by this transformation."
                )

                failed_size = (~passed).sum()
                slice_size = len(passed)

                issue = Issue(
                    model,
                    dataset,
                    group=self._issue_group,
                    level=issue_level,
                    transformation_fn=transformation_fn,
                    description=desc,
                    features=[feature],
                    meta={
                        "feature": feature,
                        "domain": f"Feature `{feature}`",
                        "deviation": f"{failed_size}/{slice_size} tested samples ({round(fail_rate * 100, 2)}%) changed prediction after perturbation",
                        "failed_size": failed_size,
                        "slice_size": slice_size,
                        "fail_rate": fail_rate,
                        "fail_rate_percent": round(fail_rate * 100, 2),
                        "metric": "Fail rate",
                        "metric_value": fail_rate,
                        "threshold": threshold,
                        "output_sentitivity": output_sensitivity,
                        "perturbed_data_slice": perturbed_data,
                        "perturbed_data_slice_predictions": perturbed_pred,
                    },
                    importance=fail_rate,
                    tests=_generate_robustness_tests,
                    taxonomy=self._taxonomy,
                    detector_name=self.__class__.__name__,
                )

                # Add examples
                examples = self._create_examples(
                    original_data,
                    original_pred,
                    perturbed_data,
                    perturbed_pred,
                    feature,
                    passed,
                    model,
                    transformation_fn,
                )
                issue.add_examples(examples)

                issues.append(issue)

        return issues

    def run(self, model: BaseModel, dataset: Dataset, features: Sequence[str]) -> Sequence[Issue]:
        """
        Runs the perturbation detector on the given model and dataset.

        Parameters
        ----------
        model: BaseModel
            The model to test.
        dataset: Dataset
            The dataset to use for testing.
        features: Sequence[str]
            The features (columns) to test.

        Returns
        -------
        Sequence[Issue]
            A list of issues found during the testing.
        """
        transformations = self.transformations or self._get_default_transformations()
        selected_features = self._select_features(dataset, features)

        logger.info(
            f"{self.__class__.__name__}: Running with transformations={[t.name for t in transformations]} "
            f"threshold={self.threshold} output_sensitivity={self.output_sensitivity} num_samples={self.num_samples}"
        )

        issues = []
        for transformation in transformations:
            issues.extend(self._detect_issues(model, dataset, transformation, selected_features))

        return [i for i in issues if i is not None]


class BaseTextPerturbationDetector(BasePerturbationDetector):
    """
    Base class for metamorphic detectors based on text transformations.
    """

    def _select_features(self, dataset: Dataset, features: Sequence[str]) -> Sequence[str]:
        # Only analyze text features
        return [
            f
            for f in features
            if dataset.column_types[f] == "text" and pd.api.types.is_string_dtype(dataset.df[f].dtype)
        ]

    @abstractmethod
    def _get_default_transformations(self) -> Sequence[TextTransformation]:
        raise NotImplementedError

    def _supports_text_generation(self) -> bool:
        return True


class BaseNumericalPerturbationDetector(BasePerturbationDetector):
    """
    Base class for metamorphic detectors based on numerical feature perturbations.
    """

    def _select_features(self, dataset: Dataset, features: Sequence[str]) -> Sequence[str]:
        # Only analyze numeric features
        return [f for f in features if dataset.column_types[f] == "numeric"]

    @abstractmethod
    def _get_default_transformations(self) -> Sequence[NumericalTransformation]:
        raise NotImplementedError

    def _supports_text_generation(self) -> bool:
        return False



import itertools


def batched(iterable, batch_size):
    """Batches an iterable into chunks of size batch_size."""
    it = iter(iterable)
    while True:
        chunk = list(itertools.islice(it, batch_size))
        if not chunk:
            break
        yield chunk



from typing import Sequence

from giskard.scanner import logger

from ...datasets.base import Dataset
from ...llm.client import get_default_client
from ...llm.client.base import ChatMessage
from ...models.base.model import BaseModel
from ..decorators import detector
from ..issues import Issue, IssueLevel, Robustness
from .base import RequirementBasedDetector

BREAKER_PROMPT = (
    "Based on the following description, does this AI agent perform a text reformulation "
    "or summarization task? Answer Y or N."
)


FAITHFULNESS_ISSUE_DESCRIPTION = (
    "Faithfulness and adherence to the instructions (the agent output is consistent with the requirements specified "
    "in the agent description and if the model is performing reformulation or summarization of an input text, the "
    "output is faithful and reliable to the input text)."
)


@detector("llm_faithfulness", tags=["faithfulness", "llm", "text_generation"])
class LLMFaithfulnessDetector(RequirementBasedDetector):
    """
    You can explicitly run this detector by adding the tag "faithfulness" in the `only` parameter of the scan method.
    """

    _issue_group = Robustness
    _issue_level = IssueLevel.MAJOR

    def get_issue_description(self) -> str:
        return FAITHFULNESS_ISSUE_DESCRIPTION

    def run(self, model: BaseModel, dataset: Dataset, features=None) -> Sequence[Issue]:
        # Let’s check whether the model is performing reformulation/summarization.
        llm_client = get_default_client()
        out = llm_client.complete(
            [
                ChatMessage(role="system", content=BREAKER_PROMPT),
                ChatMessage(role="user", content="Agent description: " + model.description),
            ],
            temperature=0.1,
            max_tokens=1,
            caller_id=self.__class__.__name__,
            seed=self.llm_seed,
        )

        if out.content.strip().upper() != "Y":
            logger.warning(
                f"{self.__class__.__name__}: Skipping faithfulness checks because the model is performing a different task."
            )
            return []

        return super().run(model, dataset, features=features)



from giskard.llm.talk.tools.base import BaseTool
from giskard.llm.talk.tools.metric import MetricTool
from giskard.llm.talk.tools.predict import PredictTool
from giskard.llm.talk.tools.scan import IssuesScannerTool
from giskard.llm.talk.tools.shap import SHAPExplanationTool

__all__ = ["BaseTool", "PredictTool", "MetricTool", "IssuesScannerTool", "SHAPExplanationTool"]



from typing import Optional, Sequence
from typing_extensions import deprecated

from logging import warning

from ..config import LLMConfigurationError
from ..errors import LLMImportError
from . import LLMClient
from .base import ChatMessage

try:
    import google.generativeai as genai
    from google.generativeai.types import ContentDict
except ImportError as err:
    raise LLMImportError(
        flavor="llm",
        msg="To use Gemini models, please install the `genai` package with `pip install google-generativeai`",
    ) from err

AUTH_ERROR_MESSAGE = (
    "Could not get Response from Gemini API. Please make sure you have configured the API key by "
    "setting GOOGLE_API_KEY in the environment."
)


def _format(messages: Sequence[ChatMessage]) -> Sequence[ContentDict]:
    system_prompts = []
    content = []

    for message in messages:
        if message.role == "system":
            system_prompts.append(message.content)

            if len(content) == 0:
                content.append(ContentDict(role="model", parts=[]))

            content[0]["parts"].insert(0, f"# System:\n{message.content}")

            continue

        role = "model" if message.role == "assistant" else "user"

        # Consecutive messages need to be grouped
        last_message = None if len(content) == 0 else content[-1]
        if last_message is not None and last_message["role"] == role:
            last_message["parts"].append(message.content)
            continue

        content.append(ContentDict(role=role, parts=[message.content]))

    return content


@deprecated("GeminiClient is deprecated: https://docs.giskard.ai/en/latest/open_source/setting_up/index.html")
class GeminiClient(LLMClient):
    def __init__(self, model: str = "gemini-pro", _client=None):
        self.model = model
        self._client = _client or genai.GenerativeModel(self.model)

    def get_config(self) -> dict:
        """Return the configuration of the LLM client."""
        return {"client_type": self.__class__.__name__, "model": self.model}

    def complete(
        self,
        messages: Sequence[ChatMessage],
        temperature: float = 1.0,
        max_tokens: Optional[int] = None,
        caller_id: Optional[str] = None,
        seed: Optional[int] = None,
        format=None,
    ) -> ChatMessage:
        if seed is not None:
            warning("Unsupported seed, ignoring.")

        if format:
            warning(f"Unsupported format '{format}', ignoring.")

        try:
            completion = self._client.generate_content(
                contents=_format(messages),
                generation_config=genai.types.GenerationConfig(temperature=temperature, max_output_tokens=max_tokens),
            )
        except RuntimeError as err:
            raise LLMConfigurationError(AUTH_ERROR_MESSAGE) from err

        self.logger.log_call(
            prompt_tokens=self._client.count_tokens([m.content for m in messages]).total_tokens,
            sampled_tokens=self._client.count_tokens(completion.text).total_tokens,
            model=self.model,
            client_class=self.__class__.__name__,
            caller_id=caller_id,
        )

        # Assuming the response structure is similar to the ChatMessage structure
        return ChatMessage(role=completion.candidates[0].content.role, content=completion.text)



from typing import Optional, Sequence, Union

import itertools

from ..utils.analytics_collector import analytics
from .knowledge_base import KnowledgeBase
from .question_generators import (
    QuestionGenerator,
    complex_questions,
    conversational_questions,
    distracting_questions,
    double_questions,
    simple_questions,
    situational_questions,
)
from .question_generators.utils import maybe_tqdm
from .testset import QATestset


def generate_testset(
    knowledge_base: KnowledgeBase,
    num_questions: Optional[int] = 120,
    question_generators: Optional[Union[QuestionGenerator, Sequence[QuestionGenerator]]] = None,
    language: Optional[str] = "en",
    agent_description: Optional[str] = "This agent is a chatbot that answers question from users.",
) -> QATestset:
    """Generate a testset from a knowledge base.

    Parameters
    ----------
    knowledge_base : KnowledgeBase
        The knowledge base to generate questions from.
    num_questions : int
        The number of questions to generate. By default 120.
    question_generators : Union[BaseQuestionModifier, Sequence[BaseQuestionModifier]]
        Question generators to use for question generation. If multiple generator are specified,
        `num_questions` will be generated with each generators. If not specified, all available question generators will be used.
    language : str, optional
        The language to use for question generation. The default is "en" to generate questions in english.
    agent_description : str, optional
        Description of the agent to be evaluated. This will be used in the prompt for question generation to get more fitting questions.

    Returns
    -------
    QATestset
        The generated test set.
    """

    if question_generators is None:
        question_generators = [
            simple_questions,
            complex_questions,
            distracting_questions,
            situational_questions,
            double_questions,
            conversational_questions,
        ]

    if not isinstance(question_generators, Sequence):
        question_generators = [question_generators]

    # Ensure topics are computed and documents populated (@TODO: remove this)
    _ = knowledge_base.topics

    # Generate questions

    # @TODO: fix this ugly way to distribute the questions across generators
    generator_num_questions = [
        num_questions // len(question_generators) + (1 if i < num_questions % len(question_generators) else 0)
        for i in range(len(question_generators))
    ]

    main_generator = itertools.chain.from_iterable(
        [
            generator.generate_questions(
                knowledge_base,
                num_questions=n,
                agent_description=agent_description,
                language=language,
            )
            for generator, n in zip(question_generators, generator_num_questions)
        ]
    )

    questions = list(maybe_tqdm(main_generator, total=num_questions, desc="Generating questions"))

    for question in questions:
        topic_id = knowledge_base[question.metadata["seed_document_id"]].topic_id
        question.metadata["topic"] = knowledge_base.topics[topic_id]

    analytics.track(
        "raget:testset-generation",
        {
            "num_questions": num_questions,
            "language": language,
            "agent_description": agent_description,
            "question_generators": [qg.__class__.__name__ for qg in question_generators],
            "knowledge_base_size": len(knowledge_base._documents),
        },
    )
    return QATestset(questions)
